from pathlib import Path

import pytest

###########################################################
# Genbank test data
###########################################################


@pytest.fixture
def gbk_testdata_dir() -> Path:
    """Genbank testdata directory"""
    return Path(__file__).parent / "testdata" / "genbank"


@pytest.fixture
def gbk_file(gbk_testdata_dir: Path) -> Path:
    """Genbank file"""
    return gbk_testdata_dir / "test.gbff"


@pytest.fixture
def gbk_bzfile(gbk_testdata_dir: Path) -> Path:
    """Genbank file (bz2 compressed)"""
    return gbk_testdata_dir / "test.gbff.bz2"


@pytest.fixture
def gbk_gzfile(gbk_testdata_dir: Path) -> Path:
    """Genbank file (gz compressed)"""
    return gbk_testdata_dir / "test.gbff.gz"


@pytest.fixture
def gbk_zipfile(gbk_testdata_dir: Path) -> Path:
    """Genbank file (zip compressed)"""
    return gbk_testdata_dir / "test.zip"


###########################################################
# GFF test data
###########################################################


@pytest.fixture
def gff_testdata_dir() -> Path:
    """GFF testdata directory"""
    return Path(__file__).parent / "testdata" / "gff"


@pytest.fixture
def gff_file(gff_testdata_dir: Path) -> Path:
    """GFF file"""
    return gff_testdata_dir / "test.gff"


@pytest.fixture
def gff_bzfile(gff_testdata_dir: Path) -> Path:
    """GFF file (bz2 compressed)"""
    return gff_testdata_dir / "test.gff.bz2"


@pytest.fixture
def gff_gzfile(gff_testdata_dir: Path) -> Path:
    """GFF file (gz compressed)"""
    return gff_testdata_dir / "test.gff.gz"


@pytest.fixture
def gff_zipfile(gff_testdata_dir: Path) -> Path:
    """GFF file (zip compressed)"""
    return gff_testdata_dir / "test.zip"



from pathlib import Path

import pytest
from pygenomeviz import Gff
from pygenomeviz.parser.gff import GffRecord

###########################################################
# Test Gff
###########################################################


def test_gff_load_default(gff_file: Path):
    """Test gff load default"""
    gff = Gff(gff_file)
    assert gff.name == "test"
    assert gff.min_range == 0
    assert gff.max_range == 60942
    assert gff.range_size == 60942


def test_gff_load_user_params(gff_file: Path):
    """Test gff load user params"""
    name, seqid, min_range, max_range = "user", "NC_000902.1", 10000, 30000
    gff = Gff(
        gff_file,
        name=name,
        target_seqid=seqid,
        min_range=min_range,
        max_range=max_range,
    )
    assert gff.name == name
    assert gff.min_range == min_range
    assert gff.max_range == max_range
    assert gff.range_size == max_range - min_range
    assert seqid == gff.seqid_list[0]


def _write_test_gff_file(
    test_gff_file: Path, delimiter: str = " ", annotation: bool = True
):
    gff_lines = (
        delimiter.join(("##sequence-region", "seqid1", "101", "1000")),
        "seqid1\tsource\tCDS\t501\t800\t.\t+\t.\tID=test1",
        delimiter.join(("##sequence-region", "seqid2", "201", "2000")),
        "seqid2\tsource\tCDS\t1001\t1300\t.\t+\t.\tID=test2",
        "seqid2\tsource\tCDS\t1401\t1820\t.\t+\t.\tID=test2",
    )
    if not annotation:
        gff_lines = [ln for ln in gff_lines if not ln.startswith("#")]
    with open(test_gff_file, "w") as f:
        f.write("\n".join(gff_lines))


def test_sequence_region_extract(tmp_path: Path):
    """Test sequence region extraction"""
    # Case1. Available sequence-region annotation line (delimiter: ' ' or '\t')
    gff_tmp_file = tmp_path / "test.gff"
    for delimiter in (" ", "\t"):
        _write_test_gff_file(gff_tmp_file, delimiter)
        gff = Gff(gff_tmp_file)
        assert gff.min_range == 100 and gff.max_range == 1000
        gff = Gff(gff_tmp_file, target_seqid="seqid2")
        assert gff.min_range == 200 and gff.max_range == 2000

    # Case2. Not available sequence-region annotation line
    _write_test_gff_file(gff_tmp_file, annotation=False)
    gff = Gff(gff_tmp_file)
    assert gff.min_range == 0 and gff.max_range == 800
    gff = Gff(gff_tmp_file, target_seqid="seqid2")
    assert gff.min_range == 0 and gff.max_range == 1820


def test_seqid_extract(tmp_path: Path):
    """Test seqid extraction"""
    gff_tmp_file = tmp_path / "test.gff"
    _write_test_gff_file(gff_tmp_file)

    # Case1. No specify target_seqid
    gff = Gff(gff_tmp_file)
    assert gff.target_seqid == "seqid1"
    assert gff.seqid_list == ["seqid1", "seqid2"]

    # Case2. Specify target_seqid
    gff = Gff(gff_tmp_file, target_seqid="seqid2")
    assert gff.target_seqid == "seqid2"
    assert gff.seqid_list == ["seqid1", "seqid2"]


def test_gff_load_seqid_exist_error(gff_file: Path):
    """Test gff load seqid exist error"""
    with pytest.raises(ValueError):
        Gff(gff_file, target_seqid="no_exists")


def test_gff_load_range_error(gff_file: Path):
    """Test gff load range error"""
    # Case1. min_range < 0
    with pytest.raises(ValueError):
        Gff(gff_file, min_range=-10000, max_range=40000)
    # Case2. min_range > max_range
    with pytest.raises(ValueError):
        Gff(gff_file, min_range=40000, max_range=10000)


def test_extract_features(gff_file: Path):
    """Test extract_features"""
    gff = Gff(gff_file)
    cds_features = gff.extract_features()
    assert len(cds_features) == 83

    no_features = gff.extract_features("no_data")
    assert len(no_features) == 0


def test_extract_features_restrict_range(gff_file: Path):
    """Test extract_features with restrict range"""
    gff = Gff(gff_file, min_range=10000, max_range=30000)
    features = gff.extract_features()
    assert len(features) == 28


def test_parse_bzfile(gff_bzfile: Path):
    """Test parse GFF file (bz2 compressed)"""
    gff = Gff(gff_bzfile)
    assert gff.name == "test"


def test_parse_gzfile(gff_gzfile: Path):
    """Test parse GFF file (gz compressed)"""
    gff = Gff(gff_gzfile)
    assert gff.name == "test"


def test_parse_zipfile(gff_zipfile: Path):
    """Test parse GFF file (zip compressed)"""
    gff = Gff(gff_zipfile)
    assert gff.name == "test"


###########################################################
# Test GffRecord
###########################################################


def test_to_gff_line():
    """Test GffRecord to_gff_line"""
    # Case1. score=None, strand=1, phase=None, attrs='ID=test'
    gff_record = GffRecord("1", "ncbi", "CDS", 10, 20, None, 1, None, {"ID": ["test"]})
    expected_gff_line = "1\tncbi\tCDS\t11\t20\t.\t+\t.\tID=test"
    assert gff_record.to_gff_line() == expected_gff_line

    # Case2. score=1e-10, strand=-1, phase=10, attrs='ID=test;Dbxref=ref1,ref2'
    attrs = {"ID": ["test"], "Dbxref": ["ref1", "ref2"]}
    gff_record = GffRecord("2", "ncbi", "mRNA", 20, 30, 1e-10, -1, 10, attrs)
    expected_gff_line = "2\tncbi\tmRNA\t21\t30\t1e-10\t-\t10\tID=test;Dbxref=ref1,ref2"
    assert gff_record.to_gff_line() == expected_gff_line


def test_parse_gff_line():
    """Test GffRecord parse_gff_line"""
    # Case1. score='.', strand='+', phase='.', attrs='ID=test'
    gff_line = "1\tncbi\tCDS\t11\t20\t.\t+\t.\tID=test"
    gff_record = GffRecord.parse_gff_line(gff_line)
    expected_gff_record = GffRecord(
        "1", "ncbi", "CDS", 10, 20, None, 1, None, {"ID": ["test"]}
    )
    assert gff_record == expected_gff_record

    # Case2. score=1e-10, strand='-', phase=10, attrs='ID=test;Dbxref=ref1,ref2'
    gff_line = "2\tncbi\tmRNA\t21\t30\t1e-10\t-\t10\tID=test;Dbxref=ref1,ref2"
    gff_record = GffRecord.parse_gff_line(gff_line)
    attrs = {"ID": ["test"], "Dbxref": ["ref1", "ref2"]}
    expected_gff_record = GffRecord("2", "ncbi", "mRNA", 20, 30, 1e-10, -1, 10, attrs)
    assert gff_record == expected_gff_record


def test_parse_and_to_gff_line():
    """Test `parse_gff_line` and `to_gff_line` results are compatible"""
    gff_line = "1\tncbi\tCDS\t11\t20\t.\t+\t.\tID=test"
    gff_record = GffRecord.parse_gff_line(gff_line)
    assert gff_line == gff_record.to_gff_line()

    gff_line = "2\tncbi\tmRNA\t21\t30\t1e-10\t-\t10\tID=test;Dbxref=ref1,ref2"
    gff_record = GffRecord.parse_gff_line(gff_line)
    assert gff_line == gff_record.to_gff_line()



from __future__ import annotations

import argparse
import csv
import os
from collections import defaultdict
from pathlib import Path

from pygenomeviz import GenomeViz, __version__
from pygenomeviz.align import AlignCoord, ProgressiveMauve
from pygenomeviz.config import LiteralTypes
from pygenomeviz.scripts import get_argparser, print_args
from pygenomeviz.utils import ColorCycler


def main():
    """Main function called from CLI"""
    # Get arguments
    args = get_args()
    print_args(args)
    # Run workflow
    run(**args.__dict__)


def run(
    # General options
    seq_files: list[str | Path],
    outdir: str | Path,
    refid: int = 0,
    format: list[str] = ["png", "html"],
    reuse: bool = False,
    # Figure appearence options
    fig_width: float = 15,
    fig_track_height: float = 1.0,
    feature_track_ratio: float = 1.0,
    link_track_ratio: float = 5.0,
    tick_track_ratio: float = 1.0,
    track_labelsize: int = 20,
    tick_labelsize: int = 15,
    normal_link_color: str = "grey",
    inverted_link_color: str = "tomato",
    align_type: LiteralTypes.ALIGN_TYPE = "center",
    tick_style: LiteralTypes.TICK_STYLE = None,
    plotstyle: LiteralTypes.PLOTSTYLE = "box",
    cmap: str = "hsv",
    curve: bool = True,
    dpi: int = 300,
) -> GenomeViz:
    """Run genome visualization workflow using progressiveMauve

    Parameters
    ----------
    seq_files : list[str | Path]
        Input genome sequence files (Genbank or Fasta format)
    outdir : str | Path
        Output directory
    format : list[str], optional
        Output image format (`png`|`jpg`|`svg`|`pdf`|`html`)
    reuse : bool, optional
        If True, reuse previous result if available
    fig_width : float, optional
        Figure width
    fig_track_height : float, optional
        Figure track height
    feature_track_ratio : float, optional
        Feature track ratio
    link_track_ratio : float, optional
        Link track ratio
    tick_track_ratio : float, optional
        Tick track ratio
    track_labelsize : int, optional
        Track label size
    tick_labelsize : int, optional
        Tick label size
    normal_link_color : str, optional
        Normal link color
    inverted_link_color : str, optional
        Inverted link color
    align_type : str, optional
        Figure tracks align type (`left`|`center`|`right`)
    tick_style : str | None, optional
        Tick style (`bar`|`axis`|`None`)
    plotstyle : str, optional
        Block box plot style (`box`|`bigbox`)
    cmap : str, optional
        Block box colormap
    curve : bool, optional
        If True, plot curved style link
    dpi : int, optional
        Figure DPI

    Returns
    -------
    gv : GenomeViz
        GenomeViz instance
    """
    # Check progressiveMauve installation
    ProgressiveMauve.check_installation()

    # Setup output contents
    outdir = Path(outdir)
    pmauve_dir = outdir / "pmauve_result"
    os.makedirs(pmauve_dir, exist_ok=True)
    align_coords_file = outdir / "align_coords.tsv"
    ColorCycler.set_cmap(cmap)

    # progressiveMauve alignment
    pmauve = ProgressiveMauve(seq_files, pmauve_dir, refid)
    if pmauve.bbone_file.exists() and reuse:
        print("Reuse previous progressiveMauve result.\n")
        align_coords = pmauve.parse_pmauve_file(pmauve.bbone_file)
        AlignCoord.write(align_coords, align_coords_file)
    else:
        print("Run progressiveMauve alignment.\n")
        align_coords = pmauve.run()
        AlignCoord.write(align_coords, align_coords_file)
    name2maxsize = get_name2maxsize(pmauve.filenames, pmauve.bbone_file)
    name2blocks = get_name2blocks(align_coords)

    # Setup GenomeViz instance
    gv = GenomeViz(
        fig_width=fig_width,
        fig_track_height=fig_track_height,
        feature_track_ratio=feature_track_ratio,
        link_track_ratio=link_track_ratio,
        tick_track_ratio=tick_track_ratio,
        align_type=align_type,
        tick_style=tick_style,
        tick_labelsize=tick_labelsize,
        plot_size_thr=0,
    )

    # Set tracks & block features
    for name in pmauve.filenames:
        maxsize = name2maxsize[name]
        track = gv.add_feature_track(name, maxsize, labelsize=track_labelsize)
        blocks = name2blocks[name]
        colors = ColorCycler.get_color_list(len(blocks))
        for block, color in zip(blocks, colors):
            start, end, strand = block
            track.add_feature(start, end, strand, plotstyle=plotstyle, facecolor=color)

    # Set links
    for ac in align_coords:
        gv.add_link(
            ac.ref_link,
            ac.query_link,
            normal_link_color,
            inverted_link_color,
            curve=curve,
        )

    # Save figure
    for fmt in format:
        output_file = outdir / f"result.{fmt}"
        if fmt == "html":
            gv.savefig_html(output_file)
        else:
            gv.savefig(output_file, dpi=dpi)
        print(f"Save {fmt} format result figure ({output_file}).")

    return gv


def get_name2maxsize(names: list[str], bbone_file: str | Path) -> dict[str, int]:
    """Get name to maxsize genome dict from bbone file

    Parameters
    ----------
    names: list[str]
        Name labels
    bbone_file : str | Path
        progressiveMauve bbone format file

    Returns
    -------
    name2maxsize : dict[str, int]
        name to maxsize dict
    """
    name2maxsize = defaultdict(int)
    with open(bbone_file) as f:
        reader = csv.reader(f, delimiter="\t")
        header_row = next(reader)
        genome_num = int(len(header_row) / 2)
        for row in reader:
            row = [abs(int(col)) for col in row]
            for i in range(genome_num):
                name, size = names[i], max(row[i * 2], row[i * 2 + 1])
                if name2maxsize[name] < size:
                    name2maxsize[name] = size
    return name2maxsize


def get_name2blocks(
    align_coords: list[AlignCoord],
) -> dict[str, list[tuple[int, int, int]]]:
    """Get name to align coord blocks from align coords

    Parameters
    ----------
    align_coords : list[AlignCoord]
        Alignment coords

    Returns
    -------
    name2blocks : dict[str, list[tuple[int, int, int]]]
        name to align coord blocks
    """
    name2blocks = defaultdict(list)
    for ac in align_coords:
        if ac.ref_block not in name2blocks[ac.ref_name]:
            name2blocks[ac.ref_name].append(ac.ref_block)
        if ac.query_block not in name2blocks[ac.query_name]:
            name2blocks[ac.query_name].append(ac.query_block)
    return name2blocks


def get_args(cli_args: list[str] | None = None) -> argparse.Namespace:
    """Get arguments

    Parameters
    ----------
    cli_args : list[str] | None, optional
        CLI arguments (Used in unittest)

    Returns
    -------
    args : argparse.Namespace
        Argument parameters
    """
    parser = get_argparser(prog_name="progressiveMauve")

    #######################################################
    # General options
    #######################################################
    general_opts = parser.add_argument_group("General Options")
    general_opts.add_argument(
        "--seq_files",
        type=Path,
        help="Input genome sequence files (Genbank or Fasta format)",
        nargs="+",
        required=True,
        metavar="IN",
    )
    general_opts.add_argument(
        "-o",
        "--outdir",
        type=Path,
        help="Output directory",
        required=True,
        metavar="OUT",
    )
    default_refid = 0
    general_opts.add_argument(
        "--refid",
        type=int,
        help=f"Reference genome index (Default: {default_refid})",
        default=default_refid,
        metavar="",
    )
    default_format = ["png", "html"]
    format_list = ["png", "jpg", "svg", "pdf", "html"]
    general_opts.add_argument(
        "--format",
        type=str,
        nargs="+",
        help="Output image format ('png'[*]|'jpg'|'svg'|'pdf'|`html`[*])",
        default=default_format,
        choices=format_list,
        metavar="",
    )
    general_opts.add_argument(
        "--reuse",
        help="Reuse previous result if available",
        action="store_true",
    )
    general_opts.add_argument(
        "-v",
        "--version",
        version=f"v{__version__}",
        help="Print version information",
        action="version",
    )
    general_opts.add_argument(
        "-h",
        "--help",
        help="Show this help message and exit",
        action="help",
    )
    #######################################################
    # Figure appearence options
    #######################################################
    fig_opts = parser.add_argument_group("Figure Appearence options")
    default_fig_width = 15
    fig_opts.add_argument(
        "--fig_width",
        type=float,
        help=f"Figure width (Default: {default_fig_width})",
        default=default_fig_width,
        metavar="",
    )
    default_fig_track_height = 1.0
    fig_opts.add_argument(
        "--fig_track_height",
        type=float,
        help=f"Figure track height (Default: {default_fig_track_height})",
        default=default_fig_track_height,
        metavar="",
    )
    default_feature_track_ratio = 1.0
    fig_opts.add_argument(
        "--feature_track_ratio",
        type=float,
        help=f"Feature track ratio (Default: {default_feature_track_ratio})",
        default=default_feature_track_ratio,
        metavar="",
    )
    default_link_track_ratio = 5.0
    fig_opts.add_argument(
        "--link_track_ratio",
        type=float,
        help=f"Link track ratio (Default: {default_link_track_ratio})",
        default=default_link_track_ratio,
        metavar="",
    )
    default_tick_track_ratio = 1.0
    fig_opts.add_argument(
        "--tick_track_ratio",
        type=float,
        help=f"Tick track ratio (Default: {default_tick_track_ratio})",
        default=default_tick_track_ratio,
        metavar="",
    )
    default_track_labelsize = 20
    fig_opts.add_argument(
        "--track_labelsize",
        type=int,
        help=f"Track label size (Default: {default_track_labelsize})",
        default=default_track_labelsize,
        metavar="",
    )
    default_tick_labelsize = 15
    fig_opts.add_argument(
        "--tick_labelsize",
        type=int,
        help=f"Tick label size (Default: {default_tick_labelsize})",
        default=default_tick_labelsize,
        metavar="",
    )
    default_normal_link_color = "grey"
    fig_opts.add_argument(
        "--normal_link_color",
        type=str,
        help=f"Normal link color (Default: '{default_normal_link_color}')",
        default=default_normal_link_color,
        metavar="",
    )
    default_inverted_link_color = "tomato"
    fig_opts.add_argument(
        "--inverted_link_color",
        type=str,
        help=f"Inverted link color (Default: '{default_inverted_link_color}')",
        default=default_inverted_link_color,
        metavar="",
    )
    default_align_type = "center"
    align_type_list = ["left", "center", "right"]
    fig_opts.add_argument(
        "--align_type",
        type=str,
        help="Figure tracks align type ('left'|'center'[*]|'right')",
        default=default_align_type,
        choices=align_type_list,
        metavar="",
    )
    default_tick_style = None
    fig_opts.add_argument(
        "--tick_style",
        type=str,
        help="Tick style ('bar'|'axis'|None[*])",
        default=default_tick_style,
        choices=["bar", "axis"],
        metavar="",
    )
    plotstyle = "box"
    plotstyle_list = ["box", "bigbox"]
    fig_opts.add_argument(
        "--plotstyle",
        type=str,
        help="Block box plot style ('box'[*]|'bigbox')",
        default=plotstyle,
        choices=plotstyle_list,
        metavar="",
    )
    default_cmap = "hsv"
    fig_opts.add_argument(
        "--cmap",
        type=str,
        help=f"Block box colormap (Default: '{default_cmap}')",
        default=default_cmap,
        metavar="",
    )
    fig_opts.add_argument(
        "--curve",
        help="Plot curved style link (Default: OFF)",
        action="store_true",
    )
    default_dpi = 300
    fig_opts.add_argument(
        "--dpi",
        type=int,
        help=f"Figure DPI (Default: {default_dpi})",
        default=default_dpi,
        metavar="",
    )
    args = parser.parse_args(cli_args)

    # Check arguments
    err_info = ""
    if len(args.seq_files) < 2:
        err_info += "--seq_files must be set at least two files.\n"
    for file in args.seq_files:
        if not file.exists():
            err_info += f"File not found '{file}'.\n"
    min_refid, max_refid = 0, len(args.seq_files) - 1
    if not min_refid <= args.refid <= max_refid:
        err_info += f"--refid must be '{min_refid} <= refid <= {max_refid}'\n"

    if err_info != "":
        parser.error("\n" + err_info)

    return args


if __name__ == "__main__":
    main()



from __future__ import annotations

import csv
import os
from dataclasses import dataclass
from io import StringIO, TextIOWrapper
from pathlib import Path
from urllib.request import urlretrieve

import matplotlib.pyplot as plt
import numpy as np
from Bio import Entrez
from matplotlib.colors import to_hex

from pygenomeviz.config import DATASETS, EXAMPLE_GFF_FILES, GITHUB_DATA_URL


def load_example_gff(
    filename: str,
    cache_dir: str | Path | None = None,
    overwrite_cache: bool = False,
) -> Path:
    """Load pygenomeviz example GFF file

    Load example GFF file from https://github.com/moshi4/pygenomeviz-data/
    and cache GFF file in local directory (Default: `~/.cache/pygenomeviz/`).

    List of example GFF filename
    - `enterobacteria_phage.gff`
    - `escherichia_coli.gff.gz`

    Parameters
    ----------
    filename : str
        GFF filename (e.g. `enterobacteria_phage.gff`)
    cache_dir : str | Path | None, optional
        Output cache directory (Default: `~/.cache/pygenomeviz/`)
    overwrite_cache : bool, optional
        If True, overwrite cached GFF file

    Returns
    -------
    gff_file : Path
        GFF file
    """
    # Check specified filename exists or not
    if filename not in EXAMPLE_GFF_FILES:
        err_msg = f"{filename=} not found."
        raise ValueError(err_msg)

    # Cache local directory
    if cache_dir is None:
        package_name = __name__.split(".")[0]
        cache_base_dir = Path.home() / ".cache" / package_name
        cache_dir = cache_base_dir / "gff"
        os.makedirs(cache_dir, exist_ok=True)
    else:
        cache_dir = Path(cache_dir)

    # Download GFF file
    gff_file_url = GITHUB_DATA_URL + f"gff/{filename}"
    gff_file_path = cache_dir / filename
    if overwrite_cache or not gff_file_path.exists():
        urlretrieve(gff_file_url, gff_file_path)

    return gff_file_path


def load_dataset(
    name: str,
    cache_dir: str | Path | None = None,
    overwrite_cache: bool = False,
) -> tuple[list[Path], list[DatasetLink]]:
    """Load pygenomeviz example genbank dataset

    Load genbank datasets from https://github.com/moshi4/pygenomeviz-data
    and cache datasets in local directory (Default: `~/.cache/pygenomeviz/`).

    List of dataset name
    - `escherichia_phage`
    - `erwinia_phage`
    - `enterobacteria_phage`
    - `escherichia_coli`
    - `mycoplasma_gallisepticum`

    Parameters
    ----------
    name : str
        Dataset name (e.g. `escherichia_phage`)

    cache_dir : str | Path | None, optional
        Output cache directory (Default: `~/.cache/pygenomeviz/`)

    overwrite_cache : bool
        If True, overwrite cached dataset

    Returns
    -------
    gbk_files, links : tuple[list[Path], list[DatasetLink]]
        Genbank files, DatasetLink list
    """
    # Check specified name dataset exists or not
    if name not in DATASETS.keys():
        err_msg = f"'{name}' dataset not found."
        raise ValueError(err_msg)

    # Dataset cache local directory
    if cache_dir is None:
        package_name = __name__.split(".")[0]
        cache_base_dir = Path.home() / ".cache" / package_name
        cache_dir = cache_base_dir / name
        os.makedirs(cache_dir, exist_ok=True)
    else:
        cache_dir = Path(cache_dir)

    # Download & cache dataset
    gbk_files: list[Path] = []
    links: list[DatasetLink] = []
    for filename in DATASETS[name]:
        file_url = GITHUB_DATA_URL + f"{name}/{filename}"
        file_path = cache_dir / filename
        if overwrite_cache or not file_path.exists():
            urlretrieve(file_url, file_path)
        if file_path.suffix in (".gb", ".gbk", ".gbff"):
            gbk_files.append(file_path)
        else:
            links = DatasetLink.load(file_path)

    return gbk_files, links


@dataclass
class DatasetLink:
    """Dataset Link DataClass (Only used in `load_dataset()` function)"""

    ref_name: str
    ref_start: int
    ref_end: int
    query_name: str
    query_start: int
    query_end: int
    identity: float

    @staticmethod
    def load(link_file: Path) -> list[DatasetLink]:
        """Load pyGenomeViz dataset link file

        Parameters
        ----------
        link_file : Path
            pyGenomeViz dataset link file

        Returns
        -------
        links : list[DatasetLink]
            DatasetLink list
        """
        with open(link_file) as f:
            reader = csv.reader(f, delimiter="\t")
            next(reader)
            links: list[DatasetLink] = []
            for row in reader:
                rname, rstart, rend = row[7], int(row[0]), int(row[1])
                qname, qstart, qend = row[8], int(row[2]), int(row[3])
                ident = float(row[6])
                links.append(
                    DatasetLink(rname, rstart, rend, qname, qstart, qend, ident)
                )
        return links


def fetch_genbank_by_accid(
    accid: str,
    gbk_outfile: str | Path | None = None,
    email: str | None = None,
) -> TextIOWrapper:
    """Fetch genbank text by 'Accession ID'

    Parameters
    ----------
    accid : str
        Accession ID
    gbk_outfile : str | Path | None, optional
        If file path is set, write fetch data to file
    email : str | None, optional
        Email address to notify download limitation (Required for bulk download)

    Returns
    -------
    TextIOWrapper
        Genbank data

    Examples
    --------
    >>> gbk_fetch_data = fetch_genbank_by_accid("JX128258.1")
    >>> gbk = Genbank(gbk_fetch_data)
    """
    Entrez.email = "" if email is None else email
    gbk_fetch_data: TextIOWrapper = Entrez.efetch(
        db="nucleotide",
        id=accid,
        rettype="gbwithparts",
        retmode="text",
    )
    if gbk_outfile is not None:
        gbk_text = gbk_fetch_data.read()
        with open(gbk_outfile, "w") as f:
            f.write(gbk_text)
        gbk_fetch_data = StringIO(gbk_text)

    return gbk_fetch_data


class ColorCycler:
    """Color Cycler Class"""

    counter = 0
    cmap = plt.get_cmap("tab10")

    def __new__(cls, n: int | None = None) -> str:
        """Get hexcolor cyclically from cmap by counter or user specified number

        `ColorCycler()` works same as `ColorCycler.get_color()` (syntactic sugar)

        Parameters
        ----------
        n : int | None, optional
            Number for color cycle. If None, counter class variable is used.

        Returns
        -------
        hexcolor : str
            Cyclic hexcolor string
        """
        return cls.get_color(n)

    @classmethod
    def reset_cycle(cls) -> None:
        """Reset cycle counter"""
        cls.counter = 0

    @classmethod
    def set_cmap(cls, name: str) -> None:
        """Set colormap (Default: `tab10`)"""
        cls.cmap = plt.get_cmap(name)
        cls.counter = 0

    @classmethod
    def get_color(cls, n: int | None = None) -> str:
        """Get hexcolor cyclically from cmap by counter or user specified number

        Parameters
        ----------
        n : int | None, optional
            Number for color cycle. If None, counter class variable is used.

        Returns
        -------
        hexcolor : str
            Cyclic hexcolor string
        """
        if n is None:
            n = cls.counter
            cls.counter += 1
        return to_hex(cls.cmap(n % cls.cmap.N), keep_alpha=True)

    @classmethod
    def get_color_list(cls, n: int | None = None) -> list[str]:
        """Get hexcolor list of colormap

        Parameters
        ----------
        n : int | None, optional
            If n is None, all(=cmap.N) hexcolors are extracted from colormap.
            If n is specified, hexcolors are extracted from n equally divided colormap.

        Returns
        -------
        hexcolor_list : list[str]
            Hexcolor list
        """
        if n is None:
            cmap_idx_list = list(range(0, cls.cmap.N))
        elif n > 0:
            cmap_idx_list = [int(i) for i in np.linspace(0, cls.cmap.N, n)]
        else:
            raise ValueError(f"{n=} is invalid number (Must be 'n > 0').")

        return [to_hex(cls.cmap(i), keep_alpha=True) for i in cmap_idx_list]



import subprocess as sp
from pathlib import Path

import pytest
from pygenomeviz.scripts.mmseqs import MMseqs
from pygenomeviz.scripts.mummer import MUMmer
from pygenomeviz.scripts.pmauve import ProgressiveMauve
from pygenomeviz.utils import load_dataset


def test_download_dataset_cli(tmp_path: Path):
    """Test download dataset CLI (pgv-download-dataset)"""
    dataset_name = "escherichia_phage"
    cmd = f"pgv-download-dataset -n {dataset_name} -o {tmp_path}"
    result = sp.run(cmd, shell=True)

    assert result.returncode == 0
    assert len(list(tmp_path.glob("*.gbk"))) == 4


def test_simpleplot_cli(tmp_path: Path):
    """Test simpleplot CLI (pgv-simpleplot)"""
    gbk_files, _ = load_dataset("enterobacteria_phage")
    gbk_files = [str(f) for f in gbk_files]

    # Test PNG,HTML image output
    for format in ("png", "html"):
        outfile = tmp_path / f"test.{format}"
        cmd = f"pgv-simpleplot --gbk_resources {' '.join((gbk_files))} -o {outfile}"
        result = sp.run(cmd, shell=True)
        assert result.returncode == 0
        assert outfile.exists()


@pytest.mark.skipif(
    condition=not MUMmer.check_installation(exit_on_false=False),
    reason="MUMmer is not installed in this environment.",
)
def test_mummer_cli(tmp_path: Path):
    """Test MUMmer CLI (pgv-mummer)"""
    gbk_files, _ = load_dataset("escherichia_phage")
    gbk_files = [str(f) for f in gbk_files]
    cmd = f"pgv-mummer --gbk_resources {' '.join(gbk_files)} -o {tmp_path} --curve"
    result = sp.run(cmd, shell=True)

    assert result.returncode == 0

    result_fig_file = tmp_path / "result.png"
    assert result_fig_file.exists()


@pytest.mark.skipif(
    condition=not MMseqs.check_installation(exit_on_false=False),
    reason="MMseqs is not installed in this environment.",
)
def test_mmseqs_cli(tmp_path: Path):
    """Test MMseqs CLI (pgv-mmseqs)"""
    gbk_files, _ = load_dataset("escherichia_phage")
    gbk_files = [str(f) for f in gbk_files]
    cmd = f"pgv-mmseqs --gbk_resources {' '.join(gbk_files)} -o {tmp_path} --curve"
    result = sp.run(cmd, shell=True)

    assert result.returncode == 0

    result_fig_file = tmp_path / "result.png"
    assert result_fig_file.exists()


@pytest.mark.skipif(
    condition=not ProgressiveMauve.check_installation(exit_on_false=False),
    reason="progressiveMauve is not installed in this environment.",
)
def test_pmauve_cli(tmp_path: Path):
    """Test progressiveMauve CLI (pgv-pmauve)"""
    gbk_files, _ = load_dataset("escherichia_phage")
    gbk_files = [str(f) for f in gbk_files]
    cmd = f"pgv-pmauve --seq_files {' '.join(gbk_files)} -o {tmp_path} --curve"
    result = sp.run(cmd, shell=True)

    assert result.returncode == 0

    result_fig_file = tmp_path / "result.png"
    assert result_fig_file.exists()



from __future__ import annotations

import bz2
import gzip
import zipfile
from collections import defaultdict
from dataclasses import dataclass
from io import TextIOWrapper
from pathlib import Path
from typing import Any, TextIO

from Bio.SeqFeature import CompoundLocation, FeatureLocation, SeqFeature


class Gff:
    """GFF Parser Class"""

    def __init__(
        self,
        gff_file: str | Path,
        name: str | None = None,
        target_seqid: str | None = None,
        min_range: int | None = None,
        max_range: int | None = None,
    ):
        """
        Parameters
        ----------
        gff_file : str | Path
            GFF file (`*.gz`, `*.bz2`, `*.zip` compressed file can be readable)
        name : str | None, optional
            name (If None, `file name` is set)
        target_seqid : str | None, optional
            Target seqid to be extracted. If None, only first seqid record is extracted.
        min_range : int | None, optional
            Min range to be extracted.
            If None, appropriate value is taken from GFF records.
        max_range : int | None, optional
            Max range to be extracted.
            If None, appropriate value is taken from GFF records.
        """
        self._gff_file = Path(gff_file)
        self._name = name
        self._records, start, end = self._parse_gff(gff_file, target_seqid)
        self.min_range = start if min_range is None else min_range
        self.max_range = end if max_range is None else max_range
        self._seq_region = (start, end)

        if not 0 <= self.min_range <= self.max_range:
            err_msg = "Range must be '0 <= min_range <= max_range' "
            err_msg += f"({self.min_range=}, {self.max_range=})"
            raise ValueError(err_msg)

    def _parse_gff(
        self,
        gff_file: str | Path,
        target_seqid: str | None,
    ) -> tuple[list[GffRecord], int, int]:
        """Parse GFF file

        Only parse target seqid record.
        If target_record is None, only parse first seqid record.

        Parameters
        ----------
        gff_file : str | Path
            GFF file
        target_seqid : str | None
            Target seqid to be extracted

        Returns
        -------
        gff_records, start, end : tuple[list[GffRecord], int, int]
            GFF record list, start, end
        """
        gff_file = Path(gff_file)
        if gff_file.suffix == ".gz":
            with gzip.open(gff_file, mode="rt") as f:
                gff_records, start, end = self._parse_gff_textio(f, target_seqid)
        elif gff_file.suffix == ".bz2":
            with bz2.open(gff_file, mode="rt") as f:
                gff_records, start, end = self._parse_gff_textio(f, target_seqid)
        elif gff_file.suffix == ".zip":
            with zipfile.ZipFile(gff_file) as zip:
                with zip.open(zip.namelist()[0]) as f:
                    io = TextIOWrapper(f)
                    gff_records, start, end = self._parse_gff_textio(io, target_seqid)
        else:
            with open(gff_file) as f:
                gff_records, start, end = self._parse_gff_textio(f, target_seqid)

        return gff_records, start, end

    def _parse_gff_textio(
        self,
        handle: TextIO,
        target_seqid: str | None = None,
    ) -> tuple[list[GffRecord], int, int]:
        """Parse GFF file TextIO

        Parameters
        ----------
        handle : TextIO
            GFF TextIO handle
        target_seqid : str | None, optional
            GFF target seqid

        Returns
        -------
        gff_records, start, end : tuple[list[GffRecord], int, int]
            GFF record list, start, end
        """
        gff_all_lines = handle.read().splitlines()
        gff_record_lines = filter(GffRecord.is_gff_line, gff_all_lines)
        gff_records = list(map(GffRecord.parse_gff_line, gff_record_lines))
        if len(gff_records) == 0:
            err_msg = f"Failed to parse '{self._gff_file}' as GFF file "
            raise ValueError(err_msg)

        seqid_list = list(dict.fromkeys([rec.seqid for rec in gff_records]))
        self._seqid_list = seqid_list
        if target_seqid is None:
            target_seqid = seqid_list[0]
        self._target_seqid = target_seqid
        if target_seqid not in seqid_list:
            err_msg = f"Not found {target_seqid=} in '{self._gff_file}'"
            raise ValueError(err_msg)
        gff_records = [rec for rec in gff_records if rec.seqid == target_seqid]

        # Try to get start-end region from '##sequence-region' annotation line
        start, end = None, None
        for line in gff_all_lines:
            if line.startswith("##sequence-region") and target_seqid in line:
                if len(line.split()) == 4:
                    start, end = line.split()[2:4]
                    start, end = int(start) - 1, int(end)
                    break
        if start is None or end is None:
            start, end = 0, max([r.end for r in gff_records])

        return gff_records, start, end

    @property
    def name(self) -> str:
        """Name"""
        if self._name is not None:
            return self._name
        if self._gff_file.suffix in (".gz", ".bz2", ".zip"):
            return self._gff_file.with_suffix("").with_suffix("").name
        else:
            return self._gff_file.with_suffix("").name

    @property
    def seq_region(self) -> tuple[int, int]:
        """GFF sequence-region start & end tuple

        If `##sequence-region` pragma is not found, seq_region=`(0, max_coords_value)`
        """
        return self._seq_region

    @property
    def records(self) -> list[GffRecord]:
        """GFF records"""
        return self._records

    @property
    def records_within_range(self) -> list[GffRecord]:
        """GFF records within min-max range"""
        target_gff_records: list[GffRecord] = []
        for rec in self.records:
            if rec.is_within_range(self.min_range, self.max_range):
                target_gff_records.append(rec)
        return target_gff_records

    @property
    def range_size(self) -> int:
        """Range size (`max_range - min_range`)"""
        return self.max_range - self.min_range

    @property
    def target_seqid(self) -> str:
        """Target seqid"""
        return self._target_seqid

    @property
    def seqid_list(self) -> list[str]:
        """seqid list"""
        return self._seqid_list

    def extract_features(
        self,
        feature_type: str = "CDS",
        target_strand: int | None = None,
        pseudogene: bool = False,
    ) -> list[SeqFeature]:
        """Extract features within min-max range

        Parameters
        ----------
        feature_type : str, optional
            Feature type (`CDS`, `gene`, `mRNA`, etc...)
        target_strand : int | None, optional
            Extract target strand
        pseudogene : bool, optional
            If True, `pseudo=`, `pseudogne=` tagged record only extract.
            If False, `pseudo=`, `pseudogene=` not tagged record only extract.

        Returns
        -------
        features : list[SeqFeature]
            Feature list
        """
        features: list[SeqFeature] = []
        for rec in self.records_within_range:
            if rec.type != feature_type:
                continue
            if target_strand is not None and rec.strand != target_strand:
                continue
            if pseudogene and ("pseudo" in rec.attrs or "pseudogene" in rec.attrs):
                features.append(rec.to_seq_feature())
            else:
                features.append(rec.to_seq_feature())

        return features

    def extract_exon_features(self, feature_type: str = "mRNA") -> list[SeqFeature]:
        """Extract exon structure features within min-max range

        Extract exons based on `parent feature` and `exon` ID-Parent relation

        Parameters
        ----------
        feature_type : str, optional
            Feature type (e.g. `mRNA`, `ncRNA` , etc...)

        Returns
        -------
        features : list[SeqFeature]
            Feature list
        """
        # Extract exon features by mRNA-exon relation
        parent_id = None
        parent_id2record: dict[str, GffRecord] = {}
        parent_id2exons: dict[str, list[GffRecord]] = defaultdict(list)
        for rec in self.records_within_range:
            if rec.type == feature_type:
                parent_id = rec.attrs.get("ID", [None])[0]
                if parent_id is None:
                    continue
                parent_id2record[parent_id] = rec
            if rec.type == "exon":
                if parent_id is not None and parent_id == rec.attrs["Parent"][0]:
                    parent_id2exons[parent_id].append(rec)

        # Set exon features
        exon_features: list[SeqFeature] = []
        for parent_id in parent_id2record.keys():
            parent_record = parent_id2record[parent_id]
            exons = parent_id2exons[parent_id]

            parent_feature = parent_record.to_seq_feature()
            if len(exons) == 1:
                exon_feature = parent_feature
            elif len(exons) >= 2:
                exons = sorted(exons, key=lambda e: e.start)
                locs = [e.to_feature_location() for e in exons]
                exon_feature = SeqFeature(
                    location=CompoundLocation(locs),
                    type=parent_feature.type,
                    id=parent_feature.id,
                    qualifiers=parent_feature.qualifiers,
                )
            else:
                # If no exon exists, skip feature extraction
                continue

            exon_features.append(exon_feature)

        return exon_features

    def __str__(self):
        return f"{self.name} ({self.min_range:,} - {self.max_range:,} bp)"


@dataclass
class GffRecord:
    """GFF Record DataClass"""

    seqid: str
    source: str
    type: str
    start: int
    end: int
    score: float | None
    strand: int
    phase: int | None
    attrs: dict[str, list[str]]

    def is_within_range(self, min_range: int, max_range: int) -> bool:
        """Check within target range or not

        Parameters
        ----------
        min_range : int
            Min range
        max_range : int
            Max range

        Returns
        -------
        check_result : bool
            Check result
        """
        if min_range <= self.start <= self.end <= max_range:
            return True
        else:
            return False

    def to_seq_feature(self) -> SeqFeature:
        """Convert GffRecord to SeqFeature"""
        return SeqFeature(
            location=self.to_feature_location(),
            type=self.type,
            id=self.attrs.get("ID", [""])[0],
            qualifiers=self.attrs,
        )

    def to_feature_location(self) -> FeatureLocation:
        """Convert GffRecord to FeatureLocation

        Returns
        -------
        feature_location : FeatureLocation
            Feature location
        """
        return FeatureLocation(self.start, self.end, self.strand)

    def to_gff_line(self) -> str:
        """Convert GffRecord to GFF record line

        Returns
        -------
        gff_line : str
            GFF record line (1-based coordinates)
        """
        return "\t".join(
            (
                self.seqid,
                self.source,
                self.type,
                str(self.start + 1),
                str(self.end),
                "." if self.score is None else str(self.score),
                "-" if self.strand == -1 else "+",
                "." if self.phase is None else str(self.phase),
                ";".join([f"{k}={','.join(v)}" for k, v in self.attrs.items()]),
            )
        )

    @staticmethod
    def is_gff_line(line: str) -> bool:
        """Check GFF record line or not

        Parameters
        ----------
        line : str
            GFF line

        Returns
        -------
        check_result : bool
            Check result
        """
        if line.startswith("#") or len(line.split("\t")) < 9:
            return False
        else:
            return True

    @staticmethod
    def parse_gff_line(gff_line: str) -> GffRecord:
        """Parse GFF record line

        Parameters
        ----------
        gff_line : str
            GFF record line (1-based coordinates)

        Returns
        -------
        gff_record : GffRecord
            GFF record (0-based coordinates)
        """
        gff_elms: list[Any] = gff_line.split("\t")[0:9]
        # start, end
        gff_elms[3], gff_elms[4] = int(gff_elms[3]) - 1, int(gff_elms[4])
        # score
        gff_elms[5] = None if gff_elms[5] in (".", "") else float(gff_elms[5])
        # strand
        if gff_elms[6] == "+":
            gff_elms[6] = 1
        elif gff_elms[6] == "-":
            gff_elms[6] = -1
        else:
            gff_elms[6] = 0
        # frame
        gff_elms[7] = None if gff_elms[7] in (".", "") else int(gff_elms[7])
        # attrs
        attr_dict: dict[str, list[str]] = {}
        attrs = str(gff_elms[8]).split(";")
        for attr in attrs:
            if attr != "" and "=" in attr:
                key, value = attr.split("=")
                attr_dict[key] = value.split(",")
        gff_elms[8] = attr_dict

        return GffRecord(*gff_elms)



from __future__ import annotations

import math
from typing import Any, Callable, Literal

from Bio.SeqFeature import SeqFeature
from matplotlib.axes import Axes

from pygenomeviz.config import LiteralTypes
from pygenomeviz.feature import ExonFeature, Feature
from pygenomeviz.link import Link
from pygenomeviz.parser import Genbank, Gff


class Track:
    """Track BaseClass"""

    def __init__(
        self,
        name: str,
        size: int,
        labelsize: int = 0,
        spines: bool = False,
        ratio: float = 1.0,
    ):
        """
        Parameters
        ----------
        name : str
            Track name
        size : int
            Track size
        labelsize : int, optional
            Track label size
        spines : bool, optional
            Display track spines
        ratio : float, optional
            Track height ratio
        """
        self.name = name
        self.size = size
        self.labelsize = labelsize
        self.spines = spines
        self.ratio = ratio
        self._offset: int | None = None
        self._ax: Axes | None = None

    @property
    def zorder(self) -> float:
        """Track zorder"""
        return 0

    @property
    def ylim(self) -> tuple[float, float]:
        """Track y min-max limit tuple"""
        return (-1.0, 1.0)

    @property
    def tick_params(self) -> dict[str, bool]:
        """Track tick parameters dict"""
        return {
            "left": False,
            "labelleft": False,
            "bottom": False,
            "labelbottom": False,
        }

    @property
    def spines_params(self) -> dict[str, bool]:
        """Spines parameters dict"""
        return {
            "left": self.spines,
            "right": self.spines,
            "top": self.spines,
            "bottom": self.spines,
        }

    @property
    def offset(self) -> int:
        """Track alignment offset"""
        if self._offset is None:
            err_msg = "Can't access offset property before calling 'plotfig' method."
            raise ValueError(err_msg)
        return self._offset

    @property
    def ax(self) -> Axes:
        """Track axes

        Can't access ax property before calling GenomeViz class `plotfig` method.

        Returns
        -------
        ax : Axes
            Track matplotlib axes object
            (Default: xlim=`(0, gv.max_track_size)`, ylim=`(-1, 1)`)
        """
        if self._ax is None:
            err_msg = "Can't access ax property before calling 'plotfig' method."
            raise ValueError(err_msg)
        return self._ax


class FeatureTrack(Track):
    """FeatureTrack Class"""

    def __init__(
        self,
        name: str,
        size: int,
        start_pos: int = 0,
        labelsize: int = 20,
        labelcolor: str = "black",
        labelmargin: float = 0.01,
        linewidth: int = 1,
        linecolor: str = "black",
        spines: bool = False,
        ratio: float = 1.0,
    ):
        """
        Parameters
        ----------
        name : str
            Track name
        size : int
            Track size
        start_pos : int, optional
            Track start position.
            Track start-end range is defined as (start_pos, start_pos + size).
        labelsize : int, optional
            Track label size
        labelcolor : str, optional
            Track label color
        labelmargin : float, optional
            Track label margin
        linewidth : int, optional
            Track line width
        linecolor : str, optional
            Track line color
        spines : bool, optional
            Display track spines
        ratio : float, optional
            Track height ratio
        """
        super().__init__(name, size, labelsize, spines, ratio)
        self._start_pos = start_pos
        self.labelcolor = labelcolor
        self.labelmargin = labelmargin
        self.linewidth = linewidth
        self.linecolor = linecolor
        self.features: list[Feature] = []
        self.subtracks: list[FeatureSubTrack] = []

        # Sublabel parameters
        self._sublabel_y: float | None = None
        self._sublabel_text: str | None = None
        self._sublabel_size: int = 10
        self._sublabel_color: str = "black"
        self._sublabel_ha: str = "left"
        self._sublabel_va: str = "top"
        self._sublabel_kws: dict[str, Any] = {}

    @property
    def start(self) -> int:
        """Track start position"""
        return self._start_pos

    @property
    def end(self) -> int:
        """Track end position"""
        return self.start + self.size

    @property
    def zorder(self) -> float:
        """Track zorder"""
        return 10

    @property
    def label_params(self) -> dict[str, Any]:
        """Label drawing parameters"""
        return {
            "s": self.name,
            "fontsize": self.labelsize,
            "color": self.labelcolor,
            "ha": "right",
            "va": "center",
            "gid": f"TrackLabel_{self.name}",
        }

    @property
    def sublabel_params(self) -> dict[str, Any]:
        """Sublabel drawing parameters"""
        return {
            "y": self._sublabel_y,
            "s": self._sublabel_text,
            "fontsize": self._sublabel_size,
            "color": self._sublabel_color,
            "ha": self._sublabel_ha,
            "va": self._sublabel_va,
            "zorder": 10,
            **self._sublabel_kws,
        }

    def _within_valid_range(
        self, start: int, end: int, raise_error_on_false: bool = False
    ) -> bool:
        """Check if start-end position within valid track range

        Parameters
        ----------
        start : int
            Start position
        end : int
            End position
        raise_error_on_false : bool, optional
            If start-end range is invalid in track, raise error

        Returns
        -------
        check_result : bool
            Check result
        """
        if self.start <= start <= self.end and self.start <= end <= self.end:
            return True
        else:
            if raise_error_on_false:
                err_msg = f"'{self.name}' track start-end range must be "
                err_msg += f"'{self.start} <= start <= end <= {self.end}' "
                err_msg += f"({start=}, {end=} is invalid)"
                raise ValueError(err_msg)
            return False

    def add_subtrack(
        self, name: str | None = None, ratio: float = 1.0, position: str = "below"
    ) -> None:
        """Add subtrack to feature track

        Parameters
        ----------
        name : str | None, optional
            Subtrack name. If None, subtrack name is automatically set.
        ratio : float, optional
            Subtrack size ratio to feature track
        position : str, optional
            Subtrack position (`above`|`below`)
        """
        subtrack_ratio = self.ratio * ratio
        subtrack_idx = len(self.subtracks) + 1
        if name is None:
            name = f"{self.name}_subtrack{subtrack_idx}"

        if name in [t.name for t in self.subtracks]:
            raise ValueError(f"subtrack.name='{name}' is already exists.")
        if position not in ("above", "below"):
            raise ValueError(f"{position=} is invalid ('above'|'below').")

        subtrack = FeatureSubTrack(
            name, self.size, self.spines, subtrack_ratio, position
        )
        self.subtracks.append(subtrack)

    def get_subtrack(self, name: str) -> FeatureSubTrack:
        """Get subtrack by name

        Parameters
        ----------
        name : str
            Target subtrack name

        Returns
        -------
        subtrack : FeatureSubTrack
            Target subtrack
        """
        name2subtrack = {t.name: t for t in self.subtracks}
        if name not in name2subtrack.keys():
            err_msg = f"subtrack.name='{name}' is not found."
            raise ValueError(err_msg)
        return name2subtrack[name]

    def set_sublabel(
        self,
        text: str | None = None,
        size: int = 15,
        color: str = "black",
        position: str = "bottom-left",
        ymargin: float = 0.2,
        sublabel_kws: dict[str, Any] | None = None,
    ) -> None:
        """Set sublabel to feature track

        Parameters
        ----------
        text : str | None, optional
            Sublabel text. If None, `{start} - {end} bp` label text is set.
        size : int, optional
            Sublabel size
        color : str, optional
            Sublabel color
        position : str, optional
            Sublabel position (`[top|bottom]-[left|center|right]`)
        ymargin : float, optional
            Sublabel y-margin
        sublabel_kws: dict[str, Any] | None, optional
            Text properties (e.g. `dict(size=12, color="red", ...)`)
            <https://matplotlib.org/stable/api/_as_gen/matplotlib.axes.Axes.text.html>
        """
        # Sublabel text setting (e.g. '0 - 1000 bp')
        default_text = f"{self.start} - {self.end} bp"
        self._sublabel_text = default_text if text is None else text
        self._sublabel_size = size
        self._sublabel_color = color
        self._sublabel_kws = {} if sublabel_kws is None else sublabel_kws
        # Sublabel position setting
        vpos, hpos = position.split("-")
        vpos_types, hpos_types = ("top", "bottom"), ("left", "center", "right")
        if vpos not in vpos_types or hpos not in hpos_types:
            err_msg = f"{position=} is invalid pattern. "
            err_msg += "position must be '[top|bottom]-[left|center|right]'"
            raise ValueError(err_msg)
        vpos2y = {"top": 1 + ymargin, "bottom": -1 - ymargin}
        vpos2va = {"top": "bottom", "bottom": "top"}
        self._sublabel_y = vpos2y[vpos]
        self._sublabel_va = vpos2va[vpos]
        self._sublabel_ha = hpos

    def add_feature(
        self,
        start: int,
        end: int,
        strand: LiteralTypes.STRAND = 1,
        label: str = "",
        labelsize: int = 15,
        labelcolor: str = "black",
        plotstyle: LiteralTypes.PLOTSTYLE = "bigarrow",
        facecolor: str = "orange",
        edgecolor: str = "black",
        linewidth: float = 0,
        labelrotation: int = 45,
        labelvpos: LiteralTypes.LABELVPOS = "strand",
        labelhpos: LiteralTypes.LABELHPOS = "center",
        labelha: LiteralTypes.LABELHA = "left",
        arrow_shaft_ratio: float = 0.5,
        size_ratio: float = 1.0,
        patch_kws: dict[str, Any] | None = None,
    ) -> None:
        """Add feature to track

        Parameters
        ----------
        start : int
            Featrue start position
        end : int
            Feature end position
        strand : int, optional
            Feature strand
        label : str, optional
            Feature label
        labelsize : int, optional
            Feature label size
        labelcolor : str, optional
            Feature label color
        plotstyle : str, optional
            Feature plot style (`bigarrow`|`arrow`|`bigbox`|`box`|`bigrbox`|`rbox`)
        facecolor : str, optional
            Feature facecolor
        edgecolor : str, optional
            Feature edgecolor
        linewidth : float, optional
            Feature edge linewidth
        labelrotation : int, optional
            Feature label rotation
        labelvpos : str, optional
            Feature label vertical position (`top`|`center`|`bottom`|`strand`)
            If 'strand' is set, 'top' or 'bottom' is auto selected by strand.
        labelhpos : str, optional
            Feature label horizontal position (`left`|`center`|`right`)
        labelha : str, optional
            Feature label horizontal alignment (`left`|`center`|`right`)
        arrow_shaft_ratio : float, optional
            Feature arrow shaft ratio
        size_ratio : float, optional
            Feature size ratio to track
        patch_kws : dict[str, Any] | None, optional
            Patch properties (e.g. `fc="red", ec="black", lw=0.5, ...`)
            <https://matplotlib.org/stable/api/_as_gen/matplotlib.patches.Patch.html>
        """
        # Check if start & end positions are within appropriate track range
        self._within_valid_range(start, end, raise_error_on_false=True)

        self.features.append(
            Feature(
                start,
                end,
                strand,
                label,
                labelsize,
                labelcolor,
                plotstyle,
                facecolor,
                edgecolor,
                linewidth,
                labelrotation,
                labelvpos,
                labelhpos,
                labelha,
                arrow_shaft_ratio,
                size_ratio,
                patch_kws,
            )
        )

    def add_exon_feature(
        self,
        exon_regions: list[tuple[int, int]],
        strand: LiteralTypes.STRAND = 1,
        label: str = "",
        labelsize: int = 15,
        labelcolor: str = "black",
        plotstyle: LiteralTypes.PLOTSTYLE = "bigarrow",
        facecolor: str = "orange",
        edgecolor: str = "black",
        linewidth: float = 0,
        labelrotation: int = 45,
        labelvpos: LiteralTypes.LABELVPOS = "strand",
        labelhpos: LiteralTypes.LABELHPOS = "center",
        labelha: LiteralTypes.LABELHA = "left",
        arrow_shaft_ratio: float = 0.5,
        size_ratio: float = 1.0,
        exon_labels: list[str] | None = None,
        exon_label_kws: dict[str, Any] | None = None,
        patch_kws: dict[str, Any] | None = None,
        intron_patch_kws: dict[str, Any] | None = None,
    ) -> None:
        """Add exon feature to track

        Parameters
        ----------
        exon_regions : list[tuple[int, int]]
            Exon feature start-end postion list
        strand : int, optional
            Feature strand
        label : str, optional
            Feature label
        labelsize : int, optional
            Feature label size
        labelcolor : str, optional
            Feature label color
        plotstyle : str, optional
            Feature plot style (`bigarrow`|`arrow`|`bigbox`|`box`|`bigrbox`|`rbox`)
        facecolor : str, optional
            Feature facecolor
        edgecolor : str, optional
            Feature edgecolor
        linewidth : float, optional
            Feature edge linewidth
        labelrotation : int, optional
            Feature label rotation
        labelvpos : str, optional
            Feature label vertical position (`top`|`center`|`bottom`|`strand`)
            If 'strand' is set, 'top' or 'bottom' is auto selected by strand.
        labelhpos : str, optional
            Feature label horizontal position (`left`|`center`|`right`)
        labelha : str, optional
            Feature label horizontal alignment (`left`|`center`|`right`)
        arrow_shaft_ratio : float, optional
            Feature arrow shaft ratio
        size_ratio : float, optional
            Feature size ratio to track
        exon_labels: list[str] | None, optional
            Exon labels. Array length must be same as `exon_regions`.
        exon_label_kws : dict[str, Any] | None, optional
            Text properties (e.g. `dict(size=12, color="red", ...)`)
            <https://matplotlib.org/stable/api/_as_gen/matplotlib.axes.Axes.text.html>
        patch_kws : dict[str, Any] | None, optional
            Exon patch properties (e.g. `dict(fc="red", ec="black", lw=0.5, ...)`)
            <https://matplotlib.org/stable/api/_as_gen/matplotlib.patches.Patch.html>
        intron_patch_kws : dict[str, Any] | None, optional
            Intron patch properties (e.g. `dict(ec="red", lw=2, ...)`)
            <https://matplotlib.org/stable/api/_as_gen/matplotlib.patches.Patch.html>
        """
        # Check if start & end positions are within appropriate track range
        for (exon_start, exon_end) in exon_regions:
            self._within_valid_range(exon_start, exon_end, raise_error_on_false=True)

        self.features.append(
            ExonFeature(
                exon_regions,
                strand,
                label,
                labelsize,
                labelcolor,
                plotstyle,
                facecolor,
                edgecolor,
                linewidth,
                labelrotation,
                labelvpos,
                labelhpos,
                labelha,
                arrow_shaft_ratio,
                size_ratio,
                exon_labels,
                exon_label_kws,
                patch_kws,
                intron_patch_kws,
            )
        )

    def add_genbank_features(
        self,
        gbk: Genbank,
        feature_type: str = "CDS",
        label_type: str | None = None,
        label_handle_func: Callable[[str], str] | None = None,
        allow_partial: bool = False,
        pseudogene: bool = False,
        labelsize: int = 15,
        labelcolor: str = "black",
        plotstyle: LiteralTypes.PLOTSTYLE = "bigarrow",
        facecolor: str = "orange",
        facecolor_handle_func: Callable[[SeqFeature], str] | None = None,
        edgecolor: str = "black",
        linewidth: float = 0,
        labelrotation: int = 45,
        labelvpos: LiteralTypes.LABELVPOS = "strand",
        labelhpos: LiteralTypes.LABELHPOS = "center",
        labelha: LiteralTypes.LABELHA = "left",
        arrow_shaft_ratio: float = 0.5,
        size_ratio: float = 1.0,
        patch_kws: dict[str, Any] | None = None,
    ) -> None:
        """Add features from genbank record

        Parameters
        ----------
        gbk : Genbank
            Genbank object
        feature_type : str, optional
            Feature type (e.g. `CDS`,`rRNA`,`tRNA`,etc...)
        label_type : str | None, optional
            Label type (e.g. `gene`,`protein_id`,`product`,etc...)
        label_handle_func : Callable[[str], str] | None, optional
            User defined function to handle label.
            Useful for filtering out unnecesary labels such as `hypothetical ~~~`,
            omitting labels with long characters, etc.
        allow_partial : bool, optional
            If True, features that are partially included in range are also extracted
        pseudogene : bool, optional
            If True and `feature_type='CDS'`, only add CDS features with
            `/pseudo` or `/pseudogene` qualifiers.
        labelsize : int, optional
            Feature label size
        labelcolor : str, optional
            Feature label color
        plotstyle : str, optional
            Feature plot style (`bigarrow`|`arrow`|`bigbox`|`box`|`bigrbox`|`rbox`)
        facecolor : str, optional
            Feature facecolor.
            If Genbank qualifiers has facecolor key (e.g. `/facecolor="red"`),
            facecolor key value is applied preferentially.
        facecolor_handle_func : Callable[[SeqFeature], str] | None, optional
            User defined function to handle feature facecolor.
        edgecolor : str, optional
            Feature edgecolor
        linewidth : float, optional
            Feature edge linewidth
        labelrotation : int, optional
            Feature label rotation
        labelvpos : str, optional
            Feature label vertical position (`top`|`center`|`bottom`|`strand`)
            If 'strand' is set, 'top' or 'bottom' is auto selected by strand.
        labelhpos : str, optional
            Feature label horizontal position (`left`|`center`|`right`)
        labelha : str, optional
            Feature label horizontal alignment (`left`|`center`|`right`)
        arrow_shaft_ratio : float, optional
            Feature arrow shaft ratio
        size_ratio : float, optional
            Feature size ratio to track
        patch_kws : dict[str, Any] | None, optional
            Patch properties (e.g. `dict(fc="red", ec="black", lw=0.5, ...)`)
            <https://matplotlib.org/stable/api/_as_gen/matplotlib.patches.Patch.html>
        """
        target_features = gbk.extract_features(
            feature_type, None, False, allow_partial, pseudogene
        )
        for feature in target_features:
            # Check if start & end positions are within appropriate track range
            start = int(str(feature.location.start))
            end = int(str(feature.location.end))
            self._within_valid_range(start, end, raise_error_on_false=True)

            # Get label value & apply handle func if exists
            label = feature.qualifiers.get(label_type, [""])[0]
            if label_handle_func is not None:
                label = label_handle_func(label)
            # Get facecolor & apply handle func if exists
            facecolor = feature.qualifiers.get("facecolor", [facecolor])[0]
            if facecolor_handle_func is not None:
                facecolor = facecolor_handle_func(feature)

            self.features.append(
                Feature(
                    start,
                    end,
                    feature.strand,
                    label,
                    labelsize,
                    labelcolor,
                    plotstyle,
                    facecolor,
                    edgecolor,
                    linewidth,
                    labelrotation,
                    labelvpos,
                    labelhpos,
                    labelha,
                    arrow_shaft_ratio,
                    size_ratio,
                    patch_kws,
                    feature,
                )
            )

    def add_gff_features(
        self,
        gff: Gff,
        feature_type: str = "CDS",
        parse_exon_intron: bool = False,
        label_type: str | None = None,
        label_handle_func: Callable[[str], str] | None = None,
        labelsize: int = 15,
        labelcolor: str = "black",
        plotstyle: LiteralTypes.PLOTSTYLE = "bigarrow",
        facecolor: str = "orange",
        facecolor_handle_func: Callable[[SeqFeature], str] | None = None,
        edgecolor: str = "black",
        linewidth: float = 0,
        labelrotation: int = 45,
        labelvpos: LiteralTypes.LABELVPOS = "strand",
        labelhpos: LiteralTypes.LABELHPOS = "center",
        labelha: LiteralTypes.LABELHA = "left",
        arrow_shaft_ratio: float = 0.5,
        size_ratio: float = 1.0,
        patch_kws: dict[str, Any] | None = None,
        intron_patch_kws: dict[str, Any] | None = None,
    ) -> None:
        """Add features from GFF record

        Parameters
        ----------
        gff : Gff
            GFF object
        feature_type : str, optional
            Feature type (e.g. `CDS`,`gene`,`mRNA`,etc...)
        parse_exon_intron : bool, optional
            If True, try to parse and add exon-intron structured feature
            (Expected to be used for eukaryote with feature_type=`mRNA`|`ncRNA`)
        label_type : str | None, optional
            Label type in attributes column (e.g. `ID`,`Name`,`product`,etc...)
        label_handle_func : Callable[[str], str] | None, optional
            User defined function to handle label.
            Useful for filtering out unnecesary labels such as `hypothetical ~~~`,
            omitting labels with long characters, etc.
        labelsize : int, optional
            Feature label size
        labelcolor : str, optional
            Feature label color
        plotstyle : str, optional
            Feature plot style (`bigarrow`|`arrow`|`bigbox`|`box`|`bigrbox`|`rbox`)
        facecolor : str, optional
            Feature facecolor.
            If GFF attributes column has facecolor tag (e.g. `facecolor=red;`),
            facecolor tag value is applied preferentially.
        facecolor_handle_func : Callable[[SeqFeature], str] | None, optional
            User defined function to handle feature facecolor.
        edgecolor : str, optional
            Feature edgecolor
        linewidth : float, optional
            Feature edge linewidth
        labelrotation : int, optional
            Feature label rotation
        labelvpos : str, optional
            Feature label vertical position (`top`|`center`|`bottom`|`strand`)
            If 'strand' is set, 'top' or 'bottom' is auto selected by strand.
        labelhpos : str, optional
            Feature label horizontal position (`left`|`center`|`right`)
        labelha : str, optional
            Feature label horizontal alignment (`left`|`center`|`right`)
        arrow_shaft_ratio : float, optional
            Feature arrow shaft ratio
        size_ratio : float, optional
            Feature size ratio to track
        patch_kws : dict[str, Any] | None, optional
            Patch properties (e.g. `dict(fc="red", ec="black", lw=0.5, ...)`)
            <https://matplotlib.org/stable/api/_as_gen/matplotlib.patches.Patch.html>
        intron_patch_kws : dict[str, Any] | None, optional
            Intron patch properties (e.g. `dict(ec="red", lw=0.5, ...)`)
            <https://matplotlib.org/stable/api/_as_gen/matplotlib.patches.Patch.html>
        """
        # Extract target features
        if parse_exon_intron:
            features = gff.extract_exon_features(feature_type)
        else:
            features = gff.extract_features(feature_type)

        # Add features in track
        for feature in features:
            # Get label value & apply handle func if exists
            label = feature.qualifiers.get(label_type, [""])[0]
            if label_handle_func is not None:
                label = label_handle_func(label)
            # Get facecolor & apply handle func if exists
            facecolor = feature.qualifiers.get("facecolor", [facecolor])[0]
            if facecolor_handle_func is not None:
                facecolor = facecolor_handle_func(feature)
            # Create feature plot property dict
            feature_kws = dict(
                strand=feature.strand,
                label=label,
                labelsize=labelsize,
                labelcolor=labelcolor,
                plotstyle=plotstyle,
                facecolor=facecolor,
                edgecolor=edgecolor,
                linewidth=linewidth,
                labelrotation=labelrotation,
                labelvpos=labelvpos,
                labelhpos=labelhpos,
                labelha=labelha,
                arrow_shaft_ratio=arrow_shaft_ratio,
                size_ratio=size_ratio,
                patch_kws=patch_kws,
                seq_feature=feature,
            )

            if parse_exon_intron:
                # Extract exon regions
                exon_regions: list[tuple[int, int]] = []
                for loc in feature.location.parts:
                    start, end = int(str(loc.start)), int(str(loc.end))
                    self._within_valid_range(start, end, raise_error_on_false=True)
                    exon_regions.append((start, end))
                # Add exon feature
                feature_kws["intron_patch_kws"] = intron_patch_kws
                self.features.append(ExonFeature(exon_regions, **feature_kws))
            else:
                start = int(str(feature.location.start))
                end = int(str(feature.location.end))
                self._within_valid_range(start, end, raise_error_on_false=True)
                # Add feature
                self.features.append(Feature(start, end, **feature_kws))

    def __str__(self):
        return f"{self.name}:{self.start}-{self.end}"


class FeatureSubTrack(Track):
    """FeatureSubTrack Class"""

    def __init__(
        self,
        name: str,
        size: int,
        spines: bool = False,
        ratio: float = 1.0,
        position: Literal["above", "below"] = "below",
    ):
        """
        Parameters
        ----------
        name : str
            Subtrack name
        size : int
            Subtrack size
        spines : bool, optional
            Display track spines
        ratio : float, optional
            Subtrack height ratio
        position : str, optional
            Subtrack position (`above`|`below`)
        """
        super().__init__(name, size, 0, spines, ratio)
        self.position = position


class LinkTrack(Track):
    """LinkTrack Class"""

    def __init__(self, name: str, spines: bool = False, ratio: float = 1.0):
        """
        Parameters
        ----------
        name : str
            Track name
        spines : bool, optional
            Display track spines
        ratio : float, optional
            Track height ratio
        """
        super().__init__(name, 0, 0, spines, ratio)
        self.links: list[Link] = []

    def add_link(self, link: Link) -> None:
        """Add link to track

        Parameters
        ----------
        link : Link
            Link between feature track
        """
        self.links.append(link)


class TickTrack(Track):
    """TickTrack Class"""

    def __init__(
        self,
        size: int,
        labelsize: int = 15,
        spines: bool = False,
        ratio: float = 1.0,
        tick_style: str = "bar",
    ):
        """
        Parameters
        ----------
        size : int
            Track size
        labelsize : int, optional
            Tick label size
        spines : bool, optional
            Display track spines
        ratio : float, optional
            Track height ratio
        tick_style : str, optional
            Tick style (`axis`|`bar`)
        """
        super().__init__("tick", size, labelsize, spines, ratio)
        self.tick_style = tick_style
        self.height_scale = 1.0

    @property
    def tick_params(self) -> dict[str, Any]:
        """Track tick parameters dict"""
        return {
            "left": False,
            "labelleft": False,
            "bottom": self.tick_style == "axis",
            "labelbottom": self.tick_style == "axis",
            "labelsize": self.labelsize,
        }

    @property
    def spines_params(self) -> dict[str, bool]:
        """Spines parameters dict"""
        return {
            "left": self.spines,
            "right": self.spines,
            "top": self.spines,
            "bottom": self.spines or self.tick_style == "axis",
        }

    @property
    def scalebar_text_params(self) -> dict[str, Any]:
        """Scalebar text parameters dict"""
        return {
            "x": self.xcenter,
            "y": self.ymin,
            "s": self.scalebar_label,
            "fontsize": self.labelsize,
            "ha": "center",
            "va": "top",
        }

    @property
    def unit(self) -> str:
        """Unit (bp, Kb, Mb, Gb)"""
        for unit, value in self.unit2base_value.items():
            if self.size >= value:
                return unit
        raise ValueError("Unexpected error.")

    @property
    def format_str(self) -> str:
        """Format string ('.0f' or '.1f')"""
        if self.size / self.base_value >= 10:
            return ".0f"
        else:
            return ".1f"

    @property
    def base_value(self) -> float:
        """Base value"""
        return self.unit2base_value[self.unit]

    @property
    def unit2base_value(self) -> dict[str, int]:
        """Unit & base value dict"""
        return {"Gb": 10**9, "Mb": 10**6, "Kb": 10**3, "bp": 1}

    @property
    def scalebar_label(self) -> str:
        """Label"""
        label = f"{self.scalebar_size / self.base_value:{self.format_str}}"
        return f"{label} {self.unit}"

    @property
    def xmin(self) -> float:
        """xmin"""
        return self.size - self.scalebar_size

    @property
    def xmax(self) -> float:
        """xmax"""
        return self.size

    @property
    def xcenter(self) -> float:
        """xcenter"""
        return (self.xmin + self.xmax) / 2

    @property
    def ymin(self) -> float:
        """ymin"""
        return self.ylim[0] - abs(self.ylim[0] * 0.1 * self.height_scale)

    @property
    def ycenter(self) -> float:
        """ycenter"""
        return self.ylim[0]

    @property
    def ymax(self) -> float:
        """ymax"""
        return self.ylim[0] + abs(self.ylim[0] * 0.1 * self.height_scale)

    @property
    def scalebar_size(self) -> float:
        """Scalebar size"""
        min_scalebar_size = self.size * 0.1
        unit = int(10 ** (len(str(int(min_scalebar_size))) - 1))
        value = math.ceil(min_scalebar_size / unit)
        steps = [1, 2, 5, 10]
        for i in range(0, len(steps) - 1):
            if steps[i] < value < steps[i + 1]:
                return steps[i + 1] * unit
        return value * unit

    def tick_formatter(self, value: float, pos: int) -> str:
        """Tick formatter

        Parameters
        ----------
        value : float
            Format target tick value
        pos : int
            Tick position (Not used for value formatting)

        Returns
        -------
        format_value : str
            Format tick value string
        """
        tick_value = value / self.base_value
        return f"{tick_value:{self.format_str}} {self.unit}"



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['pygenomeviz', 'pygenomeviz.parser', 'pygenomeviz.scripts']

package_data = \
{'': ['*'],
 'pygenomeviz': ['viewer/*',
                 'viewer/assets/*',
                 'viewer/assets/lib/*',
                 'viewer/assets/lib/images/*']}

install_requires = \
['biopython>=1.79,<2.0', 'matplotlib>=3.5.2,<4.0.0', 'numpy>=1.21,<2.0']

entry_points = \
{'console_scripts': ['pgv-download-dataset = '
                     'pygenomeviz.scripts.download_dataset:main',
                     'pgv-mmseqs = pygenomeviz.scripts.mmseqs:main',
                     'pgv-mummer = pygenomeviz.scripts.mummer:main',
                     'pgv-pmauve = pygenomeviz.scripts.pmauve:main',
                     'pgv-simpleplot = pygenomeviz.scripts.simpleplot:main']}

setup_kwargs = {
    'name': 'pygenomeviz',
    'version': '0.3.2',
    'description': 'A genome visualization python package for comparative genomics',
    'long_description': '# pyGenomeViz\n\n![Python3](https://img.shields.io/badge/Language-Python3-steelblue)\n![OS](https://img.shields.io/badge/OS-_Windows_|_Mac_|_Linux-steelblue)\n![License](https://img.shields.io/badge/License-MIT-steelblue)\n[![Latest PyPI version](https://img.shields.io/pypi/v/pygenomeviz.svg)](https://pypi.python.org/pypi/pygenomeviz)\n[![Bioconda](https://img.shields.io/conda/vn/bioconda/pygenomeviz.svg?color=green)](https://anaconda.org/bioconda/pygenomeviz)\n[![CI](https://github.com/moshi4/pyGenomeViz/actions/workflows/ci.yml/badge.svg)](https://github.com/moshi4/pyGenomeViz/actions/workflows/ci.yml)\n\n## Table of contents\n\n- [Overview](#overview)\n- [Installation](#installation)\n- [API Examples](#api-examples)\n- [CLI Examples](#cli-examples)\n- [Interactive HTML Viewer](#interactive-html-viewer)\n- [Inspiration](#inspiration)\n\n## Overview\n\npyGenomeViz is a genome visualization python package for comparative genomics implemented based on matplotlib.\nThis package is developed for the purpose of easily and beautifully plotting genomic\nfeatures and sequence similarity comparison links between multiple genomes.\nIt supports genome visualization of Genbank/GFF format file and can be saved figure in various formats (JPG/PNG/SVG/PDF/HTML).\nUser can use pyGenomeViz for interactive genome visualization figure plotting on jupyter notebook,\nor automatic genome visualization figure plotting in genome analysis scripts/pipelines.\n\nFor more information, please see full documentation [here](https://moshi4.github.io/pyGenomeViz/).\n\n![pygenomeviz_gallery.png](https://raw.githubusercontent.com/moshi4/pyGenomeViz/main/docs/images/pygenomeviz_gallery.png)  \n**Fig.1 pyGenomeViz example plot gallery**\n\n## Installation\n\n`Python 3.8 or later` is required for installation.\n\n**Install PyPI package:**\n\n    pip install pygenomeviz\n\n**Install bioconda package:**\n\n    conda install -c conda-forge -c bioconda pygenomeviz\n\n**Use Docker ([Image Registry](https://github.com/moshi4/pyGenomeViz/pkgs/container/pygenomeviz)):**\n\n    docker pull ghcr.io/moshi4/pygenomeviz:latest\n    docker run -it --rm ghcr.io/moshi4/pygenomeviz:latest pgv-pmauve -h\n\n## API Examples\n\nJupyter notebooks containing code examples below is available [here](https://moshi4.github.io/pyGenomeViz/getting_started/).\n\n### Basic Example\n\n#### Single Track\n\n```python\nfrom pygenomeviz import GenomeViz\n\nname, genome_size = "Tutorial 01", 5000\ncds_list = ((100, 900, -1), (1100, 1300, 1), (1350, 1500, 1), (1520, 1700, 1), (1900, 2200, -1), (2500, 2700, 1), (2700, 2800, -1), (2850, 3000, -1), (3100, 3500, 1), (3600, 3800, -1), (3900, 4200, -1), (4300, 4700, -1), (4800, 4850, 1))\n\ngv = GenomeViz()\ntrack = gv.add_feature_track(name, genome_size)\nfor idx, cds in enumerate(cds_list, 1):\n    start, end, strand = cds\n    track.add_feature(start, end, strand, label=f"CDS{idx:02d}")\n\ngv.savefig("example01.png")\n```\n\n![example01.png](https://raw.githubusercontent.com/moshi4/pyGenomeViz/main/docs/images/example01.png)\n\n#### Multiple Tracks & Links\n\n```python\nfrom pygenomeviz import GenomeViz\n\ngenome_list = (\n    {"name": "genome 01", "size": 1000, "cds_list": ((150, 300, 1), (500, 700, -1), (750, 950, 1))},\n    {"name": "genome 02", "size": 1300, "cds_list": ((50, 200, 1), (350, 450, 1), (700, 900, -1), (950, 1150, -1))},\n    {"name": "genome 03", "size": 1200, "cds_list": ((150, 300, 1), (350, 450, -1), (500, 700, -1), (700, 900, -1))},\n)\n\ngv = GenomeViz(tick_style="axis")\nfor genome in genome_list:\n    name, size, cds_list = genome["name"], genome["size"], genome["cds_list"]\n    track = gv.add_feature_track(name, size)\n    for idx, cds in enumerate(cds_list, 1):\n        start, end, strand = cds\n        track.add_feature(start, end, strand, label=f"gene{idx:02d}", linewidth=1, labelrotation=0, labelvpos="top", labelhpos="center", labelha="center")\n\n# Add links between "genome 01" and "genome 02"\ngv.add_link(("genome 01", 150, 300), ("genome 02", 50, 200))\ngv.add_link(("genome 01", 700, 500), ("genome 02", 900, 700))\ngv.add_link(("genome 01", 750, 950), ("genome 02", 1150, 950))\n# Add links between "genome 02" and "genome 03"\ngv.add_link(("genome 02", 50, 200), ("genome 03", 150, 300), normal_color="skyblue", inverted_color="lime", curve=True)\ngv.add_link(("genome 02", 350, 450), ("genome 03", 450, 350), normal_color="skyblue", inverted_color="lime", curve=True)\ngv.add_link(("genome 02", 900, 700), ("genome 03", 700, 500), normal_color="skyblue", inverted_color="lime", curve=True)\ngv.add_link(("genome 03", 900, 700), ("genome 02", 1150, 950), normal_color="skyblue", inverted_color="lime", curve=True)\n\ngv.savefig("example02.png")\n```\n\n![example02.png](https://raw.githubusercontent.com/moshi4/pyGenomeViz/main/docs/images/example02.png)\n\n#### Exon Features\n\n```python\nfrom pygenomeviz import GenomeViz\n\nexon_regions1 = [(0, 210), (300, 480), (590, 800), (850, 1000), (1030, 1300)]\nexon_regions2 = [(1500, 1710), (2000, 2480), (2590, 2800)]\nexon_regions3 = [(3000, 3300), (3400, 3690), (3800, 4100), (4200, 4620)]\n\ngv = GenomeViz()\ntrack = gv.add_feature_track(name=f"Exon Features", size=5000)\ntrack.add_exon_feature(exon_regions1, strand=1, plotstyle="box", label="box", labelrotation=0, labelha="center")\ntrack.add_exon_feature(exon_regions2, strand=-1, plotstyle="arrow", label="arrow", labelrotation=0, labelha="center", facecolor="darkgrey", intron_patch_kws={"ec": "red"})\n\nexon_labels = [f"exon{i+1}" for i in range(len(exon_regions3))]\ntrack.add_exon_feature(exon_regions3, strand=1, plotstyle="bigarrow", label="bigarrow", facecolor="lime", linewidth=1, exon_labels=exon_labels, labelrotation=0, labelha="center", exon_label_kws={"y": 0, "va": "center", "color": "blue"})\n\ngv.savefig("example03.png")\n```\n\n![example03.png](https://raw.githubusercontent.com/moshi4/pyGenomeViz/main/docs/images/example03.png)\n\n### Practical Example\n\n#### Add Features from Genbank file\n\n```python\nfrom pygenomeviz import Genbank, GenomeViz, load_dataset\n\ngbk_files, _ = load_dataset("enterobacteria_phage")\ngbk = Genbank(gbk_files[0])\n\ngv = GenomeViz()\ntrack = gv.add_feature_track(gbk.name, gbk.range_size)\ntrack.add_genbank_features(gbk)\n\ngv.savefig("example04.png")\n```\n\n![example04.png](https://raw.githubusercontent.com/moshi4/pyGenomeViz/main/docs/images/example04.png)\n\n#### Add Features from GFF file\n\n```python\nfrom pygenomeviz import Gff, GenomeViz, load_example_gff\n\ngff_file = load_example_gff("enterobacteria_phage.gff")\ngff = Gff(gff_file, min_range=5000, max_range=25000)\n\ngv = GenomeViz(fig_track_height=0.7, tick_track_ratio=0.5, tick_style="bar")\ntrack = gv.add_feature_track(gff.name, size=gff.range_size, start_pos=gff.min_range)\ntrack.add_gff_features(gff, plotstyle="arrow", facecolor="tomato")\ntrack.set_sublabel()\n\ngv.savefig("example05.png")\n```\n\n![example05.png](https://raw.githubusercontent.com/moshi4/pyGenomeViz/main/docs/images/example05.png)\n\n#### Multiple Tracks & Links from Genbank files\n\n```python\nfrom pygenomeviz import Genbank, GenomeViz, load_dataset\n\ngv = GenomeViz(\n    fig_track_height=0.7,\n    feature_track_ratio=0.2,\n    tick_track_ratio=0.4,\n    tick_style="bar",\n    align_type="center",\n)\n\ngbk_files, links = load_dataset("escherichia_phage")\nfor gbk_file in gbk_files:\n    gbk = Genbank(gbk_file)\n    track = gv.add_feature_track(gbk.name, gbk.range_size)\n    track.add_genbank_features(gbk, facecolor="limegreen", linewidth=0.5, arrow_shaft_ratio=1.0)\n\nfor link in links:\n    link_data1 = (link.ref_name, link.ref_start, link.ref_end)\n    link_data2 = (link.query_name, link.query_start, link.query_end)\n    gv.add_link(link_data1, link_data2, v=link.identity, curve=True)\n\ngv.savefig("example06.png")\n```\n\n![example06.png](https://raw.githubusercontent.com/moshi4/pyGenomeViz/main/docs/images/example06.png)\n\n### Customization Tips\n\nSince pyGenomeViz is implemented based on matplotlib, users can easily customize\nthe figure in the manner of matplotlib. Here are some tips for figure customization.\n\n#### Customization Tips 01\n\n- Add `GC Content` & `GC skew` subtrack\n- Add annotation label & fillbox\n- Add colorbar for links identity\n\n<details>\n<summary>Code</summary>\n\n```python\nfrom pygenomeviz import Genbank, GenomeViz, load_dataset\n\ngv = GenomeViz(\n    fig_width=12,\n    fig_track_height=0.7,\n    feature_track_ratio=0.5,\n    tick_track_ratio=0.3,\n    tick_style="axis",\n    tick_labelsize=10,\n)\n\ngbk_files, links = load_dataset("erwinia_phage")\ngbk_list = [Genbank(gbk_file) for gbk_file in gbk_files]\nfor gbk in gbk_list:\n    track = gv.add_feature_track(gbk.name, gbk.range_size, labelsize=15)\n    track.add_genbank_features(gbk, plotstyle="arrow")\n\nmin_identity = int(min(link.identity for link in links))\nfor link in links:\n    link_data1 = (link.ref_name, link.ref_start, link.ref_end)\n    link_data2 = (link.query_name, link.query_start, link.query_end)\n    gv.add_link(link_data1, link_data2, v=link.identity, vmin=min_identity)\n\n# Add subtracks to top track for plotting \'GC content\' & \'GC skew\'\ngv.top_track.add_subtrack(ratio=0.7, name="gc_content")\ngv.top_track.add_subtrack(ratio=0.7, name="gc_skew")\n\nfig = gv.plotfig()\n\n# Add label annotation to top track\ntop_track = gv.top_track  # or, gv.get_track("MT939486") or gv.get_tracks()[0]\nlabel, start, end = "Inverted", 310000 + top_track.offset, 358000 + top_track.offset\ncenter = int((start + end) / 2)\ntop_track.ax.hlines(1.5, start, end, colors="red", linewidth=1, linestyles="dashed", clip_on=False)\ntop_track.ax.text(center, 2.0, label, fontsize=12, color="red", ha="center", va="bottom")\n\n# Add fillbox to top track\nx, y = (start, start, end, end), (1, -1, -1, 1)\ntop_track.ax.fill(x, y, fc="lime", linewidth=0, alpha=0.1, zorder=-10)\n\n# Plot GC content for top track\npos_list, gc_content_list = gbk_list[0].calc_gc_content()\npos_list += gv.top_track.offset  # Offset is required if align_type is not \'left\'\ngc_content_ax = gv.top_track.subtracks[0].ax\ngc_content_ax.set_ylim(bottom=0, top=max(gc_content_list))\ngc_content_ax.fill_between(pos_list, gc_content_list, alpha=0.2, color="blue")\ngc_content_ax.text(gv.top_track.offset, max(gc_content_list) / 2, "GC(%) ", ha="right", va="center", color="blue")\n\n# Plot GC skew for top track\npos_list, gc_skew_list = gbk_list[0].calc_gc_skew()\npos_list += gv.top_track.offset  # Offset is required if align_type is not \'left\'\ngc_skew_abs_max = max(abs(gc_skew_list))\ngc_skew_ax = gv.top_track.subtracks[1].ax\ngc_skew_ax.set_ylim(bottom=-gc_skew_abs_max, top=gc_skew_abs_max)\ngc_skew_ax.fill_between(pos_list, gc_skew_list, alpha=0.2, color="red")\ngc_skew_ax.text(gv.top_track.offset, 0, "GC skew ", ha="right", va="center", color="red")\n\n# Set coloarbar for link\ngv.set_colorbar(fig, vmin=min_identity)\n\nfig.savefig("example07.png")\n```\n\n</details>\n\n![example07.png](https://raw.githubusercontent.com/moshi4/pyGenomeViz/main/docs/images/example07.png)\n\n#### Customization Tips 02\n\n- Add legends\n- Add colorbar for links identity\n\n<details>\n<summary>Code</summary>\n\n```python\nfrom matplotlib.lines import Line2D\nfrom matplotlib.patches import Patch\n\nfrom pygenomeviz import Genbank, GenomeViz, load_dataset\n\ngv = GenomeViz(\n    fig_width=10,\n    fig_track_height=0.5,\n    feature_track_ratio=0.5,\n    tick_track_ratio=0.3,\n    align_type="center",\n    tick_style="bar",\n    tick_labelsize=10,\n)\n\ngbk_files, links = load_dataset("enterobacteria_phage")\nfor idx, gbk_file in enumerate(gbk_files):\n    gbk = Genbank(gbk_file)\n    track = gv.add_feature_track(gbk.name, gbk.range_size, labelsize=10)\n    track.add_genbank_features(\n        gbk,\n        label_type="product" if idx == 0 else None,  # Labeling only top track\n        label_handle_func=lambda s: "" if s.startswith("hypothetical") else s,  # Ignore \'hypothetical ~~~\' label\n        labelsize=8,\n        labelvpos="top",\n        facecolor="skyblue",\n        linewidth=0.5,\n    )\n\nnormal_color, inverted_color, alpha = "chocolate", "limegreen", 0.5\nmin_identity = int(min(link.identity for link in links))\nfor link in links:\n    link_data1 = (link.ref_name, link.ref_start, link.ref_end)\n    link_data2 = (link.query_name, link.query_start, link.query_end)\n    gv.add_link(link_data1, link_data2, normal_color, inverted_color, alpha, v=link.identity, vmin=min_identity, curve=True)\n\nfig = gv.plotfig()\n\n# Add Legends (Maybe there is a better way)\nhandles = [\n    Line2D([], [], marker=">", color="skyblue", label="CDS", ms=10, ls="none"),\n    Patch(color=normal_color, label="Normal Link"),\n    Patch(color=inverted_color, label="Inverted Link"),\n]\nfig.legend(handles=handles, bbox_to_anchor=(1, 1))\n\n# Set colorbar for link\ngv.set_colorbar(fig, bar_colors=[normal_color, inverted_color], alpha=alpha, vmin=min_identity, bar_label="Identity", bar_labelsize=10)\n\nfig.savefig("example08.png")\n```\n\n</details>\n\n![example08.png](https://raw.githubusercontent.com/moshi4/pyGenomeViz/main/docs/images/example08.png)\n\n## CLI Examples\n\npyGenomeViz provides CLI workflow for visualization of genome alignment or\nreciprocal best-hit CDS search results with `MUMmer` or `MMseqs` or `progressiveMauve`.\nEach CLI workflow requires the installation of additional dependent tools to run.\n\n### MUMmer CLI Workflow Example\n\nSee [pgv-mummer document](https://moshi4.github.io/pyGenomeViz/cli-docs/pgv-mummer/) for details.\n\nDownload example dataset: `pgv-download-dataset -n erwinia_phage`\n\n> :warning: MUMmer must be installed in advance to run\n\n    pgv-mummer --gbk_resources MT939486.gbk MT939487.gbk MT939488.gbk LT960552.gbk \\\n               -o mummer_example --tick_style axis --align_type left --feature_plotstyle arrow\n\n![mummer_example.png](https://raw.githubusercontent.com/moshi4/pyGenomeViz/main/docs/images/mummer_example1.png)  \n\n### MMseqs CLI Workflow Example\n\nSee [pgv-mmseqs document](https://moshi4.github.io/pyGenomeViz/cli-docs/pgv-mmseqs/) for details.\n\nDownload example dataset: `pgv-download-dataset -n enterobacteria_phage`\n\n> :warning: MMseqs must be installed in advance to run\n\n    pgv-mmseqs --gbk_resources NC_019724.gbk NC_024783.gbk NC_016566.gbk NC_013600.gbk NC_031081.gbk NC_028901.gbk \\\n               -o mmseqs_example --fig_track_height 0.7 --feature_linewidth 0.3 --tick_style bar --curve \\\n               --normal_link_color chocolate --inverted_link_color limegreen --feature_color skyblue\n\n![mmseqs_example.png](https://raw.githubusercontent.com/moshi4/pyGenomeViz/main/docs/images/mmseqs_example3.png)  \n\n### progressiveMauve CLI Workflow Example\n\nSee [pgv-pmauve document](https://moshi4.github.io/pyGenomeViz/cli-docs/pgv-pmauve/) for details.\n\nDownload example dataset: `pgv-download-dataset -n escherichia_coli`\n\n> :warning: progressiveMauve must be installed in advance to run\n\n    pgv-pmauve --seq_files NC_000913.gbk NC_002695.gbk NC_011751.gbk NC_011750.gbk \\\n               -o pmauve_example --tick_style bar\n\n![pmauve_example.png](https://raw.githubusercontent.com/moshi4/pyGenomeViz/main/docs/images/pmauve_example1.png)  \n\n## Interactive HTML Viewer\n\npyGenomeViz implements HTML file output functionality for interactive data visualization.\nIn API, HTML file can be output using `savefig_html` method. In CLI, user can select HTML file output option.\nAs shown below, data tooltip display, pan/zoom, object color change, text change, etc are available in HTML viewer\n([Demo Page](https://moshi4.github.io/pyGenomeViz/images/pgv-viewer-demo.html)).\n\n![pgv-viewer-demo.gif](https://raw.githubusercontent.com/moshi4/pyGenomeViz/main/docs/images/pgv-viewer-demo.gif)\n\n## Inspiration\n\npyGenomeViz was inspired by\n\n- [GenomeDiagram (BioPython)](https://github.com/biopython/biopython)\n- [Easyfig](http://mjsull.github.io/Easyfig/)\n- [genoplotR](https://genoplotr.r-forge.r-project.org/)\n- [gggenomes](https://github.com/thackl/gggenomes)\n',
    'author': 'moshi4',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://moshi4.github.io/pyGenomeViz/',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

import csv
import itertools
import multiprocessing as mp
import os
import shutil
import subprocess as sp
import sys
import tempfile
from abc import ABCMeta, abstractmethod
from dataclasses import astuple, dataclass
from pathlib import Path
from typing import ClassVar, Literal

from pygenomeviz import Genbank


class AlignToolBase(metaclass=ABCMeta):
    """Alignment Tool Abstract Base Class"""

    # Program name
    NAME: str = ""
    # Required binary names to run
    BINARIES: list[str] = []

    @abstractmethod
    def run(self) -> list[AlignCoord]:
        """Run genome alignment

        Returns
        -------
        align_coords : list[AlignCoord]
            Genome alignment coord list
        """
        raise NotImplementedError

    @property
    def max_process_num(self) -> int:
        """Max process number"""
        cpu_num = os.cpu_count()
        return 1 if cpu_num is None or cpu_num == 1 else cpu_num - 1

    @classmethod
    def check_installation(cls, exit_on_false: bool = True) -> bool:
        """Check required binaries installation

        Parameters
        ----------
        exit_on_error : bool
            If True and check result is False, system exit

        Returns
        -------
        result : bool
            Check result
        """
        is_installed = True
        for required_binary in cls.BINARIES:
            if not shutil.which(required_binary):
                is_installed = False

        if not is_installed and exit_on_false:
            err_msg = f"ERROR: Genome alignment by {cls.NAME} is not available "
            err_msg += f"in this environment. Please check {cls.NAME} installation."
            print(err_msg)
            sys.exit(1)

        return is_installed


class MUMmer(AlignToolBase):
    """MUMmer Alignment Class"""

    NAME = "MUMmer"
    BINARIES = ["nucmer", "promer", "delta-filter", "show-coords"]

    def __init__(
        self,
        gbk_resources: list[Genbank],
        outdir: str | Path,
        seqtype: Literal["protein", "nucleotide"] = "protein",
        maptype: Literal["many-to-many", "one-to-one"] = "many-to-many",
        process_num: int | None = None,
    ):
        """
        Parameters
        ----------
        gbk_resources : list[Genbank]
            Genbank objects
        outdir : str | Path
            Output directory
        seqtype : str, optional
            Sequence type (`nucleotide`|`protein`)
        maptype : str, optional
            Alignment map type (`one-to-one`|`many-to-many`)
        process_num : int | None, optional
            Use processor number (Default: `'Max Processor' - 1`)
        """
        self.check_installation()

        self.gbk_resources = gbk_resources
        self.outdir = Path(outdir)
        self.seqtype = seqtype
        self.maptype = maptype
        self.process_num = self.max_process_num if process_num is None else process_num

    @property
    def genome_num(self) -> int:
        """Input genome fasta file number"""
        return len(self.gbk_resources)

    @property
    def _genome_fasta_files(self) -> list[Path]:
        """Genome fasta file list"""
        genome_fasta_files = []
        for gr in self.gbk_resources:
            suffix = "_reverse.fna" if gr.reverse else ".fna"
            filename = f"{gr.name}_{gr.min_range}-{gr.max_range}{suffix}"
            genome_fasta_file = self.outdir / filename
            if not genome_fasta_file.exists():
                gr.write_genome_fasta(genome_fasta_file)
            genome_fasta_files.append(genome_fasta_file)
        return genome_fasta_files

    @property
    def _align_binary(self) -> str:
        """Genome alignment binary name (`nucleotide='nucmer'`|`protein='promer'`)"""
        if self.seqtype == "nucleotide":
            return "nucmer"
        elif self.seqtype == "protein":
            return "promer"
        else:
            raise ValueError(f"Invalid seqtype '{self.seqtype}'")

    @property
    def _map_option(self) -> str:
        """Genome alignment mapping option (`one-to-one='-1'`|`many-to-many='-m'`)"""
        if self.maptype == "one-to-one":
            return "-1"
        elif self.maptype == "many-to-many":
            return "-m"
        else:
            raise ValueError(f"Invalid maptype '{self.maptype}'")

    def run(self) -> list[AlignCoord]:
        """Run genome alignment

        Returns
        -------
        align_coords : list[AlignCoord]
            Genome alignment coord list
        """
        # Prepare data for run MUMmer with multiprocessing
        mp_data_list: list[tuple[Path, Path, int]] = []
        for idx in range(0, self.genome_num - 1):
            fa_file1 = self._genome_fasta_files[idx]
            fa_file2 = self._genome_fasta_files[idx + 1]
            mp_data_list.append((fa_file1, fa_file2, idx))

        # Run MUMmer with multiprocessing
        with mp.Pool(processes=self.process_num) as p:
            results = p.starmap(self._run_mummer, mp_data_list)

        align_coords = list(itertools.chain.from_iterable(results))

        # Add min_range to start, end coordinates
        gbk_name2min_range = {gbk.name: gbk.min_range for gbk in self.gbk_resources}
        for ac in align_coords:
            ac.ref_start = ac.ref_start + gbk_name2min_range[ac.ref_name]
            ac.ref_end = ac.ref_end + gbk_name2min_range[ac.ref_name]
            ac.query_start = ac.query_start + gbk_name2min_range[ac.query_name]
            ac.query_end = ac.query_end + gbk_name2min_range[ac.query_name]

        return align_coords

    def _run_mummer(self, fa_file1: Path, fa_file2: Path, idx: int) -> list[AlignCoord]:
        """Run MUMmer function for multiprocessing

        Parameters
        ----------
        fa_file1 : Path
            Genome fasta file 01
        fa_file2 : Path
            Genome fasta file 02
        idx : int
            Multiprocessing index

        Returns
        -------
        align_coords : list[AlignCoord]
            Align coord list
        """
        # Run genome alignment using nucmer or promer
        prefix = self.outdir / f"out{idx}"
        delta_file = prefix.with_suffix(".delta")
        cmd = f"{self._align_binary} --mum {fa_file1} {fa_file2} --prefix={prefix}"
        _ = sp.run(cmd, shell=True, capture_output=True, text=True)

        # Run delta-filter to map 'one-to-one' or 'many-to-many' relation
        filter_delta_file = self.outdir / f"filter_out{idx}.delta"
        cmd = f"delta-filter {self._map_option} {delta_file} > {filter_delta_file}"
        _ = sp.run(cmd, shell=True, capture_output=True, text=True)

        # Run show-coords to extract alingment coords
        # -H: no header, -T: tab-delimited format, -r: sort by ref-id,
        # -k: knockout alignments that overlap in different frame (promer only)
        coords_file = self.outdir / f"coords{idx}.tsv"
        cmd = f"show-coords -H -T -r -k {filter_delta_file} > {coords_file}"
        _ = sp.run(cmd, shell=True, capture_output=True, text=True)

        align_coords = self.parse_coords_file(coords_file, self.seqtype)

        # Delete work files
        for work_file in (delta_file, filter_delta_file, coords_file):
            os.unlink(work_file)

        return align_coords

    @staticmethod
    def parse_coords_file(
        coords_tsv_file: str | Path,
        seqtype,
    ) -> list[AlignCoord]:
        """Parse MUMmer(nucmer|promer) output coords result file

        Parameters
        ----------
        coords_tsv_file : str | Path
            MUMmer align coords file

        Returns
        -------
        align_coords : list[AlignCoord]
            Align coord list
        """
        align_coords = []
        with open(coords_tsv_file) as f:
            reader = csv.reader(f, delimiter="\t")
            for row in reader:
                # Check read file contents & extract required row values
                if seqtype == "nucleotide":
                    if len(row) != 9:
                        err_msg = f"Invalid nucmer coords file '{coords_tsv_file}'!!"
                        raise ValueError(err_msg)
                elif seqtype == "protein":
                    if len(row) != 13:
                        err_msg = f"Invalid promer coords file '{coords_tsv_file}'!!"
                        raise ValueError(err_msg)
                    row = row[0:7] + row[11:13]
                else:
                    raise ValueError(f"Invalid seqtype '{seqtype}'!!")

                # Convert to correct value type
                typed_row = []
                for idx, val in enumerate(row):
                    if 0 <= idx <= 5:
                        typed_row.append(int(val))
                    elif idx == 6:
                        typed_row.append(float(val))
                    else:
                        typed_row.append(str(val))

                align_coords.append(AlignCoord(*typed_row))

        return align_coords


class MMseqs(AlignToolBase):
    """MMseqs Alignment Class"""

    NAME = "MMseqs"
    BINARIES = ["mmseqs"]

    def __init__(
        self,
        gbk_resources: list[str | Path] | list[Genbank],
        outdir: str | Path,
        identity: float = 0,
        evalue: float = 1e-3,
        process_num: int | None = None,
    ):
        """
        Parameters
        ----------
        gbk_resources : list[str | Path] | list[Genbank]
            Genome sequence genbank files or Genbank objects
        outdir : str | Path
            Output directory
        identity : float, optional
            Identity threshold
        evalue : float, optional
            E-value threshold
        process_num : int | None, optional
            Use processor number (Default: `'Max Processor' - 1`)
        """
        self.check_installation()

        self.gbk_list: list[Genbank] = []
        for gr in gbk_resources:
            if isinstance(gr, Genbank):
                self.gbk_list.append(gr)
            else:
                self.gbk_list.append(Genbank(gr))
        self.outdir = Path(outdir)
        self.identity = identity
        self.evalue = evalue
        self.process_num = self.max_process_num if process_num is None else process_num

        os.makedirs(self.outdir, exist_ok=True)

    def run(self) -> list[AlignCoord]:
        """Run genome alignment

        Returns
        -------
        align_coords : list[AlignCoord]
            Genome alignment coord list
        """
        # Make CDS fasta files from Genbank files
        cds_fasta_files: list[Path] = []
        for gbk in self.gbk_list:
            cds_fasta_file = self.outdir / (gbk.name + ".faa")
            gbk.write_cds_fasta(cds_fasta_file)
            cds_fasta_files.append(cds_fasta_file)

        align_coords = []
        for idx in range(0, len(self.gbk_list) - 1):
            fa_file1, fa_file2 = cds_fasta_files[idx], cds_fasta_files[idx + 1]

            with tempfile.TemporaryDirectory() as tmpdir:
                name1 = fa_file1.with_suffix("").name
                name2 = fa_file2.with_suffix("").name
                rbh_result_file = self.outdir / f"{idx+1:02d}_{name1}-{name2}_rbh.tsv"
                cmd = f"mmseqs easy-rbh {fa_file1} {fa_file2} {rbh_result_file} "
                cmd += f"{tmpdir} --threads {self.process_num} -e {self.evalue} -v 0"
                print(f"# {idx+1:02d}: {name1}-{name2} RBH search\n$ {cmd}\n")
                sp.run(cmd, shell=True)

                align_coords.extend(
                    self.parse_rbh_result(rbh_result_file, name1, name2)
                )

        return align_coords

    def parse_rbh_result(
        self,
        rbh_result_file: str | Path,
        ref_name: str,
        query_name: str,
    ) -> list[AlignCoord]:
        """Parse MMseqs RBH result

        Parameters
        ----------
        rbh_result_file : str | Path
            MMseqs RBH result file
        ref_name : str
            Reference name
        query_name : str
            Query name

        Returns
        -------
        align_coords : list[AlignCoord]
            Align Coords
        """
        align_coords = []
        dup_check_list = []
        with open(rbh_result_file) as f:
            reader = csv.reader(f, delimiter="\t")
            for row in reader:
                # Reference
                ref_start, ref_end, ref_strand = [
                    int(d) for d in row[0].split("|")[1].split("_")
                ]
                ref_length = ref_end - ref_start + 1
                if ref_strand == -1:
                    ref_start, ref_end = ref_end, ref_start
                # Query
                query_start, query_end, query_strand = [
                    int(d) for d in row[1].split("|")[1].split("_")
                ]
                query_length = query_end - query_start + 1
                if query_strand == -1:
                    query_start, query_end = query_end, query_start
                # Identity (%) [e.g. 0.95215 -> 95.21]
                identity = int(float(row[2]) * 10000) / 100

                # Check duplication
                ref_key = f"{ref_name} {ref_start} {ref_end} {ref_strand}"
                query_key = f"{query_name} {query_start} {query_end} {query_strand}"
                if ref_key in dup_check_list:
                    continue
                if query_key in dup_check_list:
                    continue
                dup_check_list.extend([ref_key, query_key])

                align_coord = AlignCoord(
                    ref_start,
                    ref_end,
                    query_start,
                    query_end,
                    ref_length,
                    query_length,
                    identity,
                    ref_name,
                    query_name,
                )
                align_coords.append(align_coord)

        return align_coords


class ProgressiveMauve(AlignToolBase):
    """progresiveMauve Alignment Class"""

    NAME = "progressiveMauve"
    BINARIES = ["progressiveMauve"]

    def __init__(
        self,
        seq_files: list[str | Path],
        outdir: str | Path,
        refid: int = 0,
    ):
        """
        Parameters
        ----------
        seq_files : list[str | Path]
            Genome sequence files (Genbank or Fasta format)
        outdir : str | Path
            Output directory
        refid : int, optional
            Reference genome index
        """
        self.check_installation()

        self.seq_files = [Path(f) for f in seq_files]
        self.outdir = Path(outdir)
        self.refid = refid

        self.xmfa_file = self.outdir / "mauve.xmfa"
        self.bbone_file = self.outdir / "mauve_bbone.tsv"

        self.seq_outdir = self.outdir / "seqfiles"
        os.makedirs(self.seq_outdir, exist_ok=True)

    @property
    def filenames(self) -> list[str]:
        """File names"""
        return [f.with_suffix("").name for f in self.seq_files]

    def run(self) -> list[AlignCoord]:
        """Run genome alignment

        Returns
        -------
        align_coords : list[AlignCoord]
            Genome alignment coord list
        """
        # Copy seqfiles to output directory
        seq_copy_files = []
        for seq_file in self.seq_files:
            seq_copy_file = self.seq_outdir / Path(seq_file).name
            # progressiveMauve cannot recognize *.gbff as genbank format
            if str(seq_copy_file).endswith(".gbff"):
                seq_copy_file = seq_copy_file.with_suffix(".gbk")
            shutil.copy(seq_file, seq_copy_file)
            seq_copy_files.append(str(seq_copy_file))

        # Run progressiveMauve
        cmd = f"progressiveMauve --output={self.xmfa_file} "
        cmd += f"--backbone-output={self.bbone_file} {' '.join(seq_copy_files)}"
        sp.run(cmd, shell=True)

        return self.parse_pmauve_file(self.bbone_file)

    def parse_pmauve_file(self, bbone_file: str | Path) -> list[AlignCoord]:
        """Parse progressiveMauve bbone file

        Parameters
        ----------
        bbone_file : str | Path
            progressiveMauve bbone format file

        Returns
        -------
        align_coords : list[AlignCoord]
            Genome alignment coord list
        """
        with open(bbone_file) as f:
            reader = csv.reader(f, delimiter="\t")
            header_row = next(reader)
            genome_num = int(len(header_row) / 2)
            rows = []
            for row in reader:
                row = [int(col) for col in row]
                ref_idx = self.refid * 2
                # Always set reference seq coordinates to positive value
                if row[ref_idx] < 0:
                    row = [col * -1 for col in row]
                # Ignore no commonly conserved regions in all genomes
                if row.count(0) >= 2:
                    continue
                # Ignore too short conserved regions (< 20bp)
                if abs(row[ref_idx] - row[ref_idx + 1]) < 20:
                    continue
                rows.append(row)
            # Sort by reference seq coordinates
            rows = sorted(rows, key=lambda row: row[ref_idx])

        align_coords = []
        for row in rows:
            for i in range(genome_num - 1):
                idx = i * 2
                # Reference start-end
                rstart, rend = row[idx], row[idx + 1]
                if rstart < 0 and rend < 0:
                    rstart, rend = abs(rend), abs(rstart)
                # Query start-end
                qstart, qend = row[idx + 2], row[idx + 3]
                if qstart < 0 and qend < 0:
                    qstart, qend = abs(qend), abs(qstart)
                rlength, qlength = abs(rend - rstart) + 1, abs(qend - qstart) + 1
                rname, qname = self.filenames[i], self.filenames[i + 1]
                align_coord = AlignCoord(
                    rstart, rend, qstart, qend, rlength, qlength, 0, rname, qname
                )
                align_coords.append(align_coord)
        return align_coords


@dataclass
class AlignCoord:
    """Alignment Coordinates DataClass"""

    ref_start: int
    ref_end: int
    query_start: int
    query_end: int
    ref_length: int
    query_length: int
    identity: float
    ref_name: str
    query_name: str
    header_list: ClassVar[list[str]] = [
        "REF_START",
        "REF_END",
        "QUERY_START",
        "QUERY_END",
        "REF_LENGTH",
        "QUERY_LENGTH",
        "IDENTITY",
        "REF_NAME",
        "QUERY_NAME",
    ]

    @property
    def ref_strand(self) -> int:
        """Reference strand"""
        return 1 if self.ref_end > self.ref_start else -1

    @property
    def ref_link(self) -> tuple[str, int, int]:
        """Reference (name, start, end) link"""
        return (self.ref_name, self.ref_start, self.ref_end)

    @property
    def ref_block(self) -> tuple[int, int, int]:
        """Reference (start, end, strand) block"""
        if self.ref_start < self.ref_end:
            return (self.ref_start, self.ref_end, self.ref_strand)
        else:
            return (self.ref_end, self.ref_start, self.ref_strand)

    @property
    def query_strand(self) -> int:
        """Query strand"""
        return 1 if self.query_end > self.query_start else -1

    @property
    def query_link(self) -> tuple[str, int, int]:
        """Query (name, start, end) link"""
        return (self.query_name, self.query_start, self.query_end)

    @property
    def query_block(self) -> tuple[int, int, int]:
        """Query (start, end, strand) block"""
        if self.query_start < self.query_end:
            return (self.query_start, self.query_end, self.query_strand)
        else:
            return (self.query_end, self.query_start, self.query_strand)

    @property
    def is_inverted(self) -> bool:
        """Check inverted or not"""
        return self.ref_strand * self.query_strand < 0

    @property
    def as_tsv_format(self) -> str:
        """TSV format text"""
        return "\t".join([str(v) for v in astuple(self)])

    @staticmethod
    def write(align_coords: list[AlignCoord], outfile: str | Path) -> None:
        """Write alignment coords as tsv format file

        Parameters
        ----------
        align_coords : list[AlignCoord]
            Alignment coords
        outfile : str | Path
            Output file path
        """
        with open(outfile, "w") as f:
            header = "\t".join(AlignCoord.header_list)
            output = "\n".join([ac.as_tsv_format for ac in align_coords])
            f.write(header + "\n" + output)

    @staticmethod
    def read(align_coords_file: str | Path) -> list[AlignCoord]:
        """Read alignment coords tsv format file

        Parameters
        ----------
        align_coords_file : str | Path
            Alignment coords tsv file

        Returns
        -------
        align_coords : list[AlignCoord]
            Alignment coords
        """
        align_coords = []
        with open(align_coords_file) as f:
            reader = csv.reader(f, delimiter="\t")
            next(reader)
            for row in reader:
                # Convert to correct value type
                typed_row = []
                for idx, val in enumerate(row):
                    if 0 <= idx <= 5:
                        typed_row.append(int(val))
                    elif idx == 6:
                        typed_row.append(float(val))
                    else:
                        typed_row.append(str(val))
                align_coords.append(AlignCoord(*typed_row))
        return align_coords

    @staticmethod
    def filter(
        align_coords: list[AlignCoord],
        min_length: int = 0,
        min_identity: float = 0.0,
    ) -> list[AlignCoord]:
        """Filter align coord list with 'length' & 'identity'

        Parameters
        ----------
        align_coords : list[AlignCoord]
            Align coord list
        min_length : int, optional
            Min length filtering threshold
        min_identity : float, optional
            Min identity filtering threshold

        Returns
        -------
        filtered_align_coords : list[AlignCoord]
            Filtered align coord list
        """
        filtered_align_coords: list[AlignCoord] = []
        for ac in align_coords:
            rlen, qlen, ident = ac.ref_length, ac.query_length, ac.identity
            if (rlen >= min_length and qlen >= min_length) and ident >= min_identity:
                filtered_align_coords.append(AlignCoord(*astuple(ac)))
        return filtered_align_coords



import pytest
from pygenomeviz.feature import ExonFeature, Feature

###########################################################
# Test Feature Class
###########################################################


def test_feature_add():
    """Test feature __add__"""
    start, end = 10, 50
    feature = Feature(start, end, strand=1)
    add_value = 10
    new_feature = feature + add_value

    assert feature.start == start
    assert feature.end == end
    assert feature.length == end - start

    assert new_feature.start == start + add_value
    assert new_feature.end == end + add_value
    assert new_feature.length == end - start


def test_feature_check_error():
    """Test feature check error case"""
    # Case1. start > end
    with pytest.raises(ValueError):
        Feature(100, 0)
    # Case2. Invalid labelvpos
    with pytest.raises(ValueError):
        Feature(0, 100, labelvpos="invalid")  # type: ignore
    # Case3. Invalid labelhpos
    with pytest.raises(ValueError):
        Feature(0, 100, labelhpos="invalid")  # type: ignore
    # Case4. Invalid labelha
    with pytest.raises(ValueError):
        Feature(0, 100, labelha="invalid")  # type: ignore
    # Case5. Invalid plotstyle
    with pytest.raises(ValueError):
        Feature(0, 100, plotstyle="invalid")  # type: ignore
    # Case6. Invalid arrow_shaft_ratio range (not 0 <= v <= 1)
    with pytest.raises(ValueError):
        Feature(0, 100, arrow_shaft_ratio=2.0)
    # Case7. Invalid size_ratio range (not 0 <= v <= 1)
    with pytest.raises(ValueError):
        Feature(0, 100, size_ratio=2.0)


###########################################################
# Test Exon Feature Class
###########################################################


def test_exon_feature_add():
    """Test exon feature __add__"""
    exon_regions = [(10, 40), (50, 110), (130, 160)]
    exon_feature = ExonFeature(exon_regions)
    add_value = 10
    new_exon_feature = exon_feature + add_value

    assert exon_feature.start == 10
    assert exon_feature.end == 160
    assert exon_feature.length == 150

    assert new_exon_feature.start == exon_feature.start + add_value
    assert new_exon_feature.end == exon_feature.end + add_value
    assert new_exon_feature.length == 150
    assert new_exon_feature.exon_regions == [(20, 50), (60, 120), (140, 170)]


def test_exon_feature_intron_regions():
    """Test exon feature intron regions"""
    exon_regions = [(10, 40), (50, 110), (130, 160)]
    exon_feature = ExonFeature(exon_regions)
    expected_intron_regions = [(40, 50), (110, 130)]
    assert exon_feature.intron_regions == expected_intron_regions


def test_exon_feature_check_error():
    """Test exon feature check error case"""
    # Case1. exon_start > exon_end
    with pytest.raises(ValueError):
        exon_regions = [(40, 10), (110, 50), (160, 130)]
        ExonFeature(exon_regions)
    # Case2. exon regions is not ascending order of genomic regions
    with pytest.raises(ValueError):
        exon_regions = [(130, 160), (50, 110), (10, 40)]
        ExonFeature(exon_regions)
